#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""Implement DataFile like classes for Various experimental rigs."""

__all__ = ["BigBlueFile", "BirgeIVFile", "MokeFile", "FmokeFile", "EasyPlotFile", "PinkLibFile"]
import re
import csv

import numpy as np

from ..compat import bytes2str
from ..core.base import string_to_type
from .. import Core
from .generic import CSVFile
from ..tools.file import FileManager


class BigBlueFile(CSVFile):
    """Extends CSVFile to load files from Nick Porter's old BigBlue code."""

    #: priority (int): is the load order for the class, smaller numbers are tried before larger numbers.
    #   .. note::
    #      Subclasses with priority<=32 should make some positive identification that they have the right
    #      file type before attempting to read data.
    priority = 64  # Also rather generic file format so make a lower priority
    #: pattern (list of str): A list of file extensions that might contain this type of file. Used to construct
    # the file load/save dialog boxes.
    patterns = ["*.dat", "*.iv", "*.rvt"]  # Recognised filename patterns

    def _load(self, filename, *args, **kargs):
        """Just call the parent class but with the right parameters set.

        Args:
            filename (string or bool): File to load. If None then the existing filename is used,
                if False, then a file dialog will be used.

        Returns:
            A copy of the itself after loading the data.
        """
        if filename is None or not filename:
            self.get_filename("r")
        else:
            self.filename = filename

        super()._load(self.filename, *args, header_line=3, data_line=7, data_delim=" ", header_delim=",")
        if np.all(np.isnan(self.data)):
            raise Core.StonerLoadError("All data was NaN in Big Blue format")
        return self


class BirgeIVFile(Core.DataFile):
    """Implements the IV File format used by the Birge Group in Michigan State University Condesned Matter Physiscs."""

    patterns = ["*.dat"]

    def _load(self, filename, *args, **kargs):
        """File loader for PinkLib.

        Args:
            filename (string or bool): File to load. If None then the existing filename is used,
                if False, then a file dialog will be used.

        Returns:
            A copy of the itself after loading the data.
        """
        if filename is None or not filename:
            self.get_filename("r")
        else:
            self.filename = filename
        ix = 0
        with FileManager(self.filename, "r", errors="ignore", encoding="utf-8") as f:  # Read filename linewise
            if not re.compile(r"\d{1,2}/\d{1,2}/\d{4}").match(f.readline()):
                raise Core.StonerLoadError("Not a BirgeIVFile as no date on first line")
            data = f.readlines()
            expected = ["Vo(-))", "Vo(+))", "Ic(+)", "Ic(-)"]
            for length, m in zip(data[-4:], expected):
                if not length.startswith(m):
                    raise Core.StonerLoadError("Not a BirgeIVFile as wrong footer line")
                key = length[: len(m)]
                val = length[len(m) :]
                if "STDEV" in val:
                    ix2 = val.index("STDEV")
                    key2 = val[ix2 : ix2 + 4 + len(key)]
                    val2 = val[ix2 + 4 + len(key) :]
                    self.metadata[key2] = string_to_type(val2.strip())
                    val = val[:ix2]
                self.metadata[key] = string_to_type(val.strip())
            for ix, line in enumerate(data):  # Scan the ough lines to get metadata
                if ":" in line:
                    parts = line.split(":")
                    self.metadata[parts[0].strip()] = string_to_type(parts[1].strip())
                elif "," in line:
                    for part in line.split(","):
                        parts = part.split(" ")
                        self.metadata[parts[0].strip()] = string_to_type(parts[1].strip())
                elif line.startswith("H "):
                    self.metadata["H"] = string_to_type(line.split(" ")[1].strip())
                else:
                    headers = [x.strip() for x in line.split(" ")]
                    break
            else:
                raise Core.StonerLoadError("Oops ran off the end of the file!")
        self.data = np.genfromtxt(filename, skip_header=ix + 2, skip_footer=4)
        self.column_headers = headers

        self.setas = "xy"
        return self


class MokeFile(Core.DataFile):
    """Class that extgends Core.DataFile to load files from the Leeds MOKE system."""

    #: priority (int): is the load order for the class, smaller numbers are tried before larger numbers.
    #   .. note::
    #      Subclasses with priority<=32 should make some positive identification that they have the right
    #      file type before attempting to read data.
    priority = 16
    #: pattern (list of str): A list of file extensions that might contain this type of file. Used to construct
    # the file load/save dialog boxes.
    patterns = ["*.dat", "*.txt"]

    def _load(self, filename, *args, **kargs):
        """Leeds  MOKE file loader routine.

        Args:
            filename (string or bool): File to load. If None then the existing filename is used,
                if False, then a file dialog will be used.

        Returns:
            A copy of the itself after loading the data.
        """
        if filename is None or not filename:
            self.get_filename("r")
        else:
            self.filename = filename
        with FileManager(self.filename, mode="rb") as f:
            line = bytes2str(f.readline()).strip()
            if line != "#Leeds CM Physics MOKE":
                raise Core.StonerLoadError("Not a Core.DataFile from the Leeds MOKE")
            while line.startswith("#") or line == "":
                parts = line.split(":")
                if len(parts) > 1:
                    key = parts[0][1:]
                    data = ":".join(parts[1:]).strip()
                    self[key] = data
                line = bytes2str(f.readline()).strip()
            column_headers = [x.strip() for x in line.split(",")]
            self.data = np.genfromtxt(f, delimiter=",")
        self.setas = "xy.de"
        self.column_headers = column_headers
        return self


class FmokeFile(Core.DataFile):
    """Extends Core.DataFile to open Fmoke Files."""

    #: priority (int): is the load order for the class, smaller numbers are tried before larger numbers.
    #   .. note::
    #      Subclasses with priority<=32 should make some positive identification that they have the right
    #      file type before attempting to read data.
    priority = 16  # Makes a positive ID check of its contents so give it priority in autoloading
    #: pattern (list of str): A list of file extensions that might contain this type of file. Used to construct
    # the file load/save dialog boxes.
    patterns = ["*.dat"]  # Recognised filename patterns

    def _load(self, filename, *args, **kargs):
        """Sheffield Focussed MOKE file loader routine.

        Args:
            filename (string or bool): File to load. If None then the existing filename is used,
                if False, then a file dialog will be used.

        Returns:
            A copy of the itself after loading the data.
        """
        if filename is None or not filename:
            self.get_filename("r")
        else:
            self.filename = filename
        with FileManager(self.filename, mode="rb") as f:
            try:
                value = [float(x.strip()) for x in bytes2str(f.readline()).split("\t")]
            except (TypeError, ValueError) as err:
                f.close()
                raise Core.StonerLoadError("Not an FMOKE file?") from err
            label = [x.strip() for x in bytes2str(f.readline()).split("\t")]
            if label[0] != "Header:":
                f.close()
                raise Core.StonerLoadError("Not a Focussed MOKE file !")
            del label[0]
            for k, v in zip(label, value):
                self.metadata[k] = v  # Create metadata from first 2 lines
            column_headers = [x.strip() for x in bytes2str(f.readline()).split("\t")]
            self.data = np.genfromtxt(f, dtype="float", delimiter="\t", invalid_raise=False)
            self.column_headers = column_headers
        return self


class EasyPlotFile(Core.DataFile):
    """A class that will extract as much as it can from an EasyPlot save File."""

    #: priority (int): is the load order for the class, smaller numbers are tried before larger numbers.
    #   .. note::
    #      Subclasses with priority<=32 should make some positive identification that they have the right
    #      file type before attempting to read data.
    priority = 32  # Fairly generic, but can do some explicit testing

    def _load(self, filename, *args, **kargs):
        """Private loader method."""
        if filename is None or not filename:
            self.get_filename("r")
        else:
            self.filename = filename

        datastart = -1
        dataend = -1

        i = 0
        with FileManager(self.filename, "r", errors="ignore", encoding="utf-8") as data:
            if "******** EasyPlot save file ********" not in data.read(1024):
                raise Core.StonerLoadError("Not an EasyPlot Save file?")
            data.seek(0)
            for i, line in enumerate(data):
                line = line.strip()
                if line == "":
                    continue
                if line[0] not in "-0123456789" and dataend < 0 <= datastart:
                    dataend = i
                if line.startswith('"') and ":" in line:
                    parts = [x.strip() for x in line.strip('"').split(":")]
                    self[parts[0]] = string_to_type(":".join(parts[1:]))
                elif line.startswith("/"):  # command
                    parts = [x.strip('"') for x in next(csv.reader([line], delimiter=" ")) if x != ""]
                    cmd = parts[0].strip("/")
                    if len(cmd) > 1:
                        cmdname = f"_{cmd}_cmd"
                        if cmdname in dir(self):  # If this command is implemented as a function run it
                            cmd = getattr(self, f"_{cmd}_cmd")
                            cmd(parts[1:])
                        else:
                            if len(parts[1:]) > 1:
                                cmd = cmd + "." + parts[1]
                                value = ",".join(parts[2:])
                            elif len(parts[1:]) == 1:
                                value = parts[1]
                            else:
                                value = True
                            self[cmd] = value
                elif line[0] in "-0123456789" and datastart < 0:  # start of data
                    datastart = i
                    if "," in line:
                        delimiter = ","
                    else:
                        delimiter = None
        if dataend < 0:
            dataend = i
        self.data = np.genfromtxt(self.filename, skip_header=datastart, skip_footer=i - dataend, delimiter=delimiter)
        if self.data.shape[1] == 2:
            self.setas = "xy"
        return self

    def _extend_columns(self, i):
        """Ensure the column headers are at least i long."""
        if len(self.column_headers) < i:
            length = len(self.column_headers)
            self.data = np.append(
                self.data, np.zeros((self.shape[0], i - length)), axis=1
            )  # Need to expand the array first
            self.column_headers.extend([f"Column {x}" for x in range(length, i)])

    def _et_cmd(self, parts):
        """Handle axis labellling command."""
        if parts[0] == "x":
            self._extend_columns(1)
            self.column_headers[0] = parts[1]
        elif parts[0] == "y":
            self._extend_columns(2)
            self.column_headers[1] = parts[1]
        elif parts[0] == "g":
            self["title"] = parts[1]

    def _td_cmd(self, parts):
        self.setas = parts[0]

    def _sa_cmd(self, parts):
        """Implement the sa (set-axis?) command."""
        if parts[0] == "l":  # Legend
            col = int(parts[2])
            self._extend_columns(col + 1)
            self.column_headers[col] = parts[1]


class PinkLibFile(Core.DataFile):
    """Extends Core.DataFile to load files from MdV's PINK library - as used by the GMR anneal rig."""

    #: priority (int): is the load order for the class, smaller numbers are tried before larger numbers.
    #   .. note::
    #      Subclasses with priority<=32 should make some positive identification that they have the right
    #      file type before attempting to read data.
    priority = 32  # reasonably generic format
    #: pattern (list of str): A list of file extensions that might contain this type of file. Used to construct
    # the file load/save dialog boxes.
    patterns = ["*.dat"]  # Recognised filename patterns

    def _load(self, filename=None, *args, **kargs):
        """File loader for PinkLib.

        Args:
            filename (string or bool): File to load. If None then the existing filename is used,
                if False, then a file dialog will be used.

        Returns:
            A copy of the itself after loading the data.
        """
        if filename is None or not filename:
            self.get_filename("r")
        else:
            self.filename = filename
        with FileManager(self.filename, "r", errors="ignore", encoding="utf-8") as f:  # Read filename linewise
            if "PINKlibrary" not in f.readline():
                raise Core.StonerLoadError("Not a PINK file")
            f = f.readlines()
            happened_before = False
            for i, line in enumerate(f):
                if line[0] != "#" and not happened_before:
                    header_line = i - 2  # -2 because there's a commented out data line
                    happened_before = True
                    continue  # want to get the metadata at the bottom of the file too
                if any(s in line for s in ("Start time", "End time", "Title")):
                    tmp = line.strip("#").split(":")
                    self.metadata[tmp[0].strip()] = ":".join(tmp[1:]).strip()
            column_headers = f[header_line].strip("#\t ").split("\t")
            data = np.genfromtxt(f, dtype="float", delimiter="\t", invalid_raise=False, comments="#")
        self.data = data[:, 0:-2]  # Deal with an errant tab at the end of each line
        self.column_headers = column_headers
        if np.all([h in column_headers for h in ("T (C)", "R (Ohm)")]):
            self.setas(x="T (C)", y="R (Ohm)")  # pylint: disable=not-callable
        return self



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""Implement DataFile classes for some generic file formats."""
__all__ = ["CSVFile", "HyperSpyFile", "KermitPNGFile", "TDMSFile"]
import csv
import contextlib
import io
import re
from collections.abc import Mapping
import sys
import logging

import PIL
import numpy as np

from ..Core import DataFile
from ..compat import str2bytes, Hyperspy_ok, hs, hsload
from ..core.exceptions import StonerLoadError
from ..tools.file import FileManager


class _refuse_log(logging.Filter):
    """Refuse to log all records."""

    def filter(self, record):
        """Do not log anything."""
        return False


@contextlib.contextmanager
def catch_sysout(*args):
    """Temporarily redirect sys.stdout and.sys.stdin."""
    stdout, stderr = sys.stdout, sys.stderr
    out = io.StringIO()
    sys.stdout, sys.stderr = out, out
    logger = logging.getLogger("hyperspy.io")
    logger.addFilter(_refuse_log)
    yield None
    logger.removeFilter(_refuse_log)
    sys.stdout, sys.stderr = stdout, stderr
    return


def _delim_detect(line):
    """Detect a delimiter in a line.

    Args:
        line(str):
            String to search for delimiters in.

    Returns:
        (str):
            Delimiter to use.

    Raises:
        StnerLoadError:
            If delimiter cannot be located.
    """
    quotes = re.compile(r"([\"\'])[^\1]*\1")
    line = quotes.sub("", line)  # Remove quoted strings first
    current = (None, len(line))
    for delim in "\t ,;":
        try:
            idx = line.index(delim)
        except ValueError:
            continue
        if idx < current[1]:
            current = (delim, idx)
    if current[0] is None:
        raise StonerLoadError("Unable to find a delimiter in the line")
    return current[0]


class CSVFile(DataFile):
    """A subclass of DataFiule for loading generic deliminated text fiules without metadata."""

    #: priority (int): is the load order for the class, smaller numbers are tried before larger numbers.
    #   .. note::
    #      Subclasses with priority<=32 should make some positive identification that they have the right
    #      file type before attempting to read data.
    priority = 128  # Rather generic file format so make it a low priority
    #: pattern (list of str): A list of file extensions that might contain this type of file. Used to construct
    # the file load/save dialog boxes.
    patterns = ["*.csv", "*.txt"]  # Recognised filename patterns

    _defaults = {"header_line": 0, "data_line": 1, "header_delim": ",", "data_delim": ","}

    mime_type = ["application/csv", "text/plain", "text/csv"]

    def _load(self, filename, *args, **kargs):
        """Load generic deliminated files.

        Args:
            filename (string or bool): File to load. If None then the existing filename is used,
                if False, then a file dialog will be used.

        Keyword Arguments:
            header_line (int): The line in the file that contains the column headers.
                If None, then column headers are automatically generated.
            data_line (int): The line on which the data starts
            data_delim (string): The delimiter used for separating data values
            header_delim (strong): The delimiter used for separating header values

        Returns:
            A copy of the current object after loading the data.
        """
        header_line = kargs.pop("header_line", self._defaults["header_line"])
        data_line = kargs.pop("data_line", self._defaults["data_line"])
        data_delim = kargs.pop("data_delim", self._defaults["data_delim"])
        header_delim = kargs.pop("header_delim", self._defaults["header_delim"])

        if filename is None or not filename:
            self.get_filename("r")
        else:
            self.filename = filename

        if data_delim is None or header_delim is None:  #
            with FileManager(self.filename, "r") as datafile:
                lines = datafile.readlines()
            if header_line is not None and header_delim is None:
                header_delim = _delim_detect(lines[header_line])
            if data_line is not None and data_delim is None:
                data_delim = _delim_detect(lines[data_line])

        with FileManager(self.filename, "r") as datafile:
            if header_line is not None:
                try:
                    for ix, line in enumerate(datafile):
                        if ix == header_line:
                            break
                    else:
                        raise StonerLoadError("Ran out of file before readching header")
                    header = line.strip()
                    column_headers = next(csv.reader(io.StringIO(header), delimiter=header_delim))
                    data = np.genfromtxt(datafile, delimiter=data_delim, skip_header=data_line - header_line)
                except (TypeError, ValueError, csv.Error, StopIteration, UnicodeDecodeError) as err:
                    raise StonerLoadError("Header and data on the same line") from err
            else:  # Generate
                try:
                    data = np.genfromtxt(datafile, delimiter=data_delim, skip_header=data_line)
                except (TypeError, ValueError) as err:
                    raise StonerLoadError("Failed to open file as CSV File") from err
                column_headers = ["Column" + str(x) for x in range(np.shape(data)[1])]

        self.data = data
        self.column_headers = column_headers
        self._kargs = kargs
        return self

    def save(self, filename=None, **kargs):
        """Override the save method to allow CSVFiles to be written out to disc (as a mininmalist output).

        Args:
            filename (string): Filename to save as (using the same rules as for the load routines)

        Keyword Arguments:
            deliminator (string): Record deliniminator (defaults to a comma)

        Returns:
            A copy of itself.
        """
        delimiter = kargs.pop("deliminator", ",")
        if filename is None:
            filename = self.filename
        if filename is None or (isinstance(filename, bool) and not filename):  # now go and ask for one
            filename = self.__file_dialog("w")
        with FileManager(filename, "w") as outfile:
            spamWriter = csv.writer(outfile, delimiter=delimiter, quotechar='"', quoting=csv.QUOTE_MINIMAL)
            i = 0
            spamWriter.writerow(self.column_headers)
            while i < self.data.shape[0]:
                spamWriter.writerow(self.data[i, :])
                i += 1
        self.filename = filename
        return self


class JustNumbersFile(CSVFile):
    """A reader format for things which are just a block of numbers with no headers or metadata."""

    priority = 256  # Rather generic file format so make it a low priority
    #: pattern (list of str): A list of file extensions that might contain this type of file. Used to construct
    # the file load/save dialog boxes.
    patterns = ["*.csv", "*.txt"]  # Recognised filename patterns

    _defaults = {"header_line": None, "data_line": 0, "header_delim": None, "data_delim": None}


class KermitPNGFile(DataFile):
    """Loads PNG files with additional metadata embedded in them and extracts as metadata."""

    #: priority (int): is the load order for the class, smaller numbers are tried before larger numbers.
    #   .. note::
    #      Subclasses with priority<=32 should make some positive identification that they have the right
    #      file type before attempting to read data.
    priority = 16  # We're checking for a the specoific PNG signature
    #: pattern (list of str): A list of file extensions that might contain this type of file. Used to construct
    # the file load/save dialog boxes.
    patterns = ["*.png"]  # Recognised filename patterns

    mime_type = ["image/png"]

    def _check_signature(self, filename):
        """Check that this is a PNG file and raie a StonerLoadError if not."""
        try:
            with FileManager(filename, "rb") as test:
                sig = test.read(8)
            sig = [x for x in sig]
            if self.debug:
                print(sig)
            if sig != [137, 80, 78, 71, 13, 10, 26, 10]:
                raise StonerLoadError("Signature mismatrch")
        except (StonerLoadError, IOError) as err:
            from traceback import format_exc

            raise StonerLoadError(f"Not a PNG file!>\n{format_exc()}") from err
        return True

    def _load(self, filename=None, *args, **kargs):
        """PNG file loader routine.

        Args:
            filename (string or bool): File to load. If None then the existing filename is used,
                if False, then a file dialog will be used.

        Returns:
            A copy of the itself after loading the data.
        """
        if filename is None or not filename:
            self.get_filename("r")
        else:
            self.filename = filename
        self._check_signature(filename)
        try:
            with PIL.Image.open(self.filename, "r") as img:
                for k in img.info:
                    self.metadata[k] = img.info[k]
                self.data = np.asarray(img)
        except IOError as err:
            raise StonerLoadError("Unable to read as a PNG file.") from err

        return self

    def save(self, filename=None, **kargs):
        """Override the save method to allow KermitPNGFiles to be written out to disc.

        Args:
            filename (string): Filename to save as (using the same rules as for the load routines)

        Keyword Arguments:
            deliminator (string): Record deliniminator (defaults to a comma)

        Returns:
            A copy of itself.
        """
        if filename is None:
            filename = self.filename
        if filename is None or (isinstance(filename, bool) and not filename):  # now go and ask for one
            filename = self.__file_dialog("w")

        metadata = PIL.PngImagePlugin.PngInfo()
        for k in self.metadata:
            parts = self.metadata.export(k).split("=")
            key = parts[0]
            val = str2bytes("=".join(parts[1:]))
            metadata.add_text(key, val)
        img = PIL.Image.fromarray(self.data)
        img.save(filename, "png", pnginfo=metadata)
        self.filename = filename
        return self


try:  # Optional tdms support
    from nptdms import TdmsFile

    class TDMSFile(DataFile):
        """First stab at writing a file that will import TDMS files."""

        #: priority (int): is the load order for the class, smaller numbers are tried before larger numbers.
        #   .. note::
        #      Subclasses with priority<=32 should make some positive identification that they have the right
        #      file type before attempting to read data.
        priority = 16  # Makes a positive ID of its file contents
        #: pattern (list of str): A list of file extensions that might contain this type of file. Used to construct
        # the file load/save dialog boxes.
        patterns = ["*.tdms"]  # Recognised filename patterns

        mime_type = ["application/octet-stream"]

        def _load(self, filename=None, *args, **kargs):
            """TDMS file loader routine.

            Args:
                filename (string or bool): File to load. If None then the existing filename is used,
                    if False, then a file dialog will be used.

            Returns:
                A copy of the itself after loading the data.
            """
            if filename is None or not filename:
                self.get_filename("r")
            else:
                self.filename = filename
            # Open the file and read the main file header and unpack into a dict
            try:
                f = TdmsFile(self.filename)
                for grp in f.groups():
                    if grp.path == "/":
                        pass  # skip the rooot group
                    elif grp.path == "/'TDI Format 1.5'":
                        tmp = DataFile(grp.as_dataframe())
                        self.data = tmp.data
                        self.column_headers = tmp.column_headers
                        self.metadata.update(grp.properties)
                    else:
                        tmp = DataFile(grp.as_dataframe())
                        self.data = tmp.data
                        self.column_headers = tmp.column_headers
            except (IOError, ValueError, TypeError, StonerLoadError) as err:
                from traceback import format_exc

                raise StonerLoadError(f"Not a TDMS File \n{format_exc()}") from err

            return self

except ImportError:
    TDMSFile = DataFile

if Hyperspy_ok:

    class HyperSpyFile(DataFile):
        """Wrap the HyperSpy file to map to DataFile."""

        priority = 64  # Makes an ID check but is quite generic

        patterns = ["*.emd", "*.dm4"]

        mime_type = ["application/x-hdf", "application/x-hdf5"]  # Really an HDF5 file

        _axes_keys = ["name", "scale", "low_index", "low_value", "high_index", "high_value"]

        def _unpack_meta(self, root, value):
            """Recursively unpack a nested dict of metadata and append keys to self.metadata."""
            if isinstance(value, Mapping):
                for item in value.keys():
                    if root != "":
                        self._unpack_meta(f"{root}.{item}", value[item])
                    else:
                        self._unpack_meta(f"{item}", value[item])
            else:
                self.metadata[root] = value

        def _unpack_axes(self, ax_manager):
            """Unpack the axes managber as metadata."""
            for ax in ax_manager.signal_axes:
                for k in self._axes_keys:
                    self.metadata[f"{ax.name}.{k}"] = getattr(ax, k)

        def _load(self, filename=None, *args, **kargs):
            """Load HyperSpy file loader routine.

            Args:
                filename (string or bool): File to load. If None then the existing filename is used,
                    if False, then a file dialog will be used.

            Returns:
                A copy of the itself after loading the data.
            """
            if filename is None or not filename:
                self.get_filename("r")
            else:
                self.filename = filename
            # Open the file and read the main file header and unpack into a dict
            try:
                with catch_sysout():
                    signal = hsload(self.filename)
                if hasattr(hs, "signals"):
                    Signal2D = hs.signals.Signal2D
                else:
                    Signal2D = hs.api.signals.Signal2D
                if not isinstance(signal, Signal2D):
                    raise StonerLoadError("Not a 2D signal object - aborting!")
            except Exception as err:  # pylint: disable=W0703 Pretty generic error catcher
                raise StonerLoadError(f"Not readable by HyperSpy error was {err}") from err
            self.data = signal.data
            self._unpack_meta("", signal.metadata.as_dictionary())
            self._unpack_axes(signal.axes_manager)

            return self

else:
    HyperSpyFile = DataFile



"""Simple XMCD Data reduction example."""

# pylint: disable=invalid-name, redefined-outer-name
import re
import numpy as np
import scipy

# Normaliser with Stoner module
from Stoner import Data, DataFolder


def helicity(f):
    """Look for either pos or neg in the first column header."""
    if "pos" in f.column_headers[0]:
        return "Pos"
    if "neg" in f.column_headers[0]:
        return "Neg"
    return "Neither"


def position(f):
    """Look for meta data named magj1yins and us this rounded to 1dp as a key."""
    if "magj1yins" in f:
        if f["magj1yins"] == -8.5:
            return "Perp"
        return "Para"
    return "None"


def temp(f):
    """Get the temperature data in bins."""
    if "sensor_temp" in f:
        return str(np.round(f["itc2"])) + "K"
    return "None"


def cycle(f):
    """If run >=100 then we're on the second cycle."""
    if f["run"] >= 49100:
        return "cycle2"
    return "cycle1"


def alt_norm(f, _, **kargs):
    """Just do the normalisation per file and report run number and normalisation constants."""
    run = f["run"]
    signal = kargs["signal"]
    lfit = kargs["lfit"]
    rfit = kargs["rfit"]
    ec = 0

    md = f.find_col(signal)
    coeffs = f.polyfit(ec, md, 1, lambda x, y: lfit[0] <= x <= lfit[1])
    linearfit = scipy.poly1d(coeffs)
    f.add_column(lambda r: r[md] - linearfit(r[ec]), "minus linear")
    highend = f.mean("minus", lambda r: rfit[0] <= r[ec] <= rfit[1])
    ret = [run]
    ret.extend(coeffs)
    ret.append(highend)
    return ret


def norm_group(pos, _, **kargs):
    """Take the drain current for each file in group and builds an analysis file and works out the mean drain."""
    if "signal" in kargs:
        signal = kargs["signal"]
    else:
        signal = "fluo"
    lfit = kargs["lfit"]
    rfit = kargs["rfit"]

    posfile = Data()
    posfile.metadata = pos[0].metadata
    posfile = posfile & pos[0].column(0)
    posfile.column_headers = ["Energy"]
    for f in pos:
        print(str(f["run"]) + str(f.find_col(signal)))
        posfile = posfile & f.column(signal)
    posfile.add_column(lambda r: np.mean(r[1:]), "mean drain")
    ec = posfile.find_col("Energy")
    md = posfile.find_col("mean drain")
    linearfit = scipy.poly1d(
        posfile.polyfit(ec, md, 1, lambda x, y: lfit[0] <= x <= lfit[1])
    )
    posfile.add_column(lambda r: r[md] - linearfit(r[ec]), "minus linear")
    highend = posfile.mean("minus", lambda r: rfit[0] <= r[ec] <= rfit[1])
    ml = posfile.find_col("minus linear")
    posfile.add_column(lambda r: r[ml] / highend, "normalised")
    if "group_key" in kargs:
        posfile[kargs["group_key"]] = pos.key
    return posfile


def asym(grp, trail, **kargs):
    """Take a group of two files for Pos and Neg helicity and calcualtes the Asymmetry ratio."""
    posfile = grp["Pos"]
    negfile = grp["Neg"]
    pdat = posfile.column("normalised")
    ndat = negfile.column("normalised")
    edat = posfile.column("Energy")
    ret = posfile.clone
    ret.column_headers = []
    ret.data = np.array([])
    ret = ret & edat & pdat & ndat
    ret.column_headers = ["Energy", "I+", "I-"]
    ret.add_column(lambda r: (r[1] - r[2]), "Asymmetry")
    ret.filename = "-".join(trail)
    if "group_key" in kargs:
        ret[kargs["group_key"]] = grp.key
    ret.title = ret.filename
    if "save" in kargs and kargs["save"]:
        ret.save()
    return ret


def collate(grp, trail, **kargs):
    """Gather all the data up again."""
    grp.sort()
    final = Data()
    final.add_column(grp[0].column("Energy"), "Energy")
    for g in grp:
        final.add_column(g.column("Asym"), g.title)
    if "group_key" in kargs:
        final[kargs["group_key"]] = grp.key
    final["path"] = trail
    if "save" in kargs and kargs["save"]:
        final.save(kargs["filename"])
    return final


# The start and end runs for this batch
startrun = 49173
endrun = 52000
# Which column are we analysing ?
signal = "fluo"
# A filename pattern that will grab the run number from the filename
pattern = re.compile(r"i10-(?P<run>\d*)\.dat")
# The Data spool directory
directory = r"C:\Data\data"
# Set the limits used on the normalisation
rfit = (660, 670)
lfit = (615, 630)

# Read the directory of data files and sort by run number
fldr = DataFolder(directory, pattern=pattern, read_means=True)
fldr.sort("run")
# Remove files outside of the run number range
fldr.filterout(lambda f: f["run"] > endrun or f["run"] < startrun)
# group the files by position, temperatures and polarisations
fldr.group([position, temp, helicity])
# Normalise the files grouped by helicity, temperature and position and produce one averaged file
# for each position, temperature and helicity
fldr.walk_groups(
    norm_group,
    group=True,
    replace_terminal=True,
    walker_args={"signal": signal, "lfit": lfit, "rfit": rfit},
)
# Calcualte the asymmetry by position and temperature and produce one file per positiona nd temeprature
# Set the walker_args to {"save":True} to save this file to disc
fldr.walk_groups(
    asym,
    group=True,
    replace_terminal=True,
    walker_args={"save": True, "group_key": "temperature"},
)
# Collate the asummetry curves into one file and save it
fldr.walk_groups(
    collate,
    group=True,
    replace_terminal=True,
    walker_args={"group_key": "Position"},
)

for f in fldr:
    f.filename = "Run 2 Compiled Fluorescence Data {}.txt".format(f["Position"])
    f.save()
# norm_data=fldr.walk_groups(alt_norm,walker_args={"rfit":(660,670),"lfit":(615,630),"signal":"fluo"})



"""Extrapolate data example."""

# pylint: disable=invalid-name
from numpy import linspace, ones_like, column_stack, exp, sqrt
from numpy.random import normal, seed
import matplotlib.pyplot as plt

from Stoner import Data
from Stoner.plot.utils import errorfill

seed(1245)  # Ensure consistent random numbers
x = linspace(1, 10, 101)
d = Data(
    column_stack(
        [
            x,
            (2 * x**2 - x + 2) + normal(size=len(x), scale=2.0),
            ones_like(x) * 2,
        ]
    ),
    setas="xye",
    column_headers=["x", "y", "errors"],
)

extra_x = linspace(8, 15, 11)

y4 = d.extrapolate(
    extra_x,
    overlap=200,
    kind=lambda x, A, C: A * exp(x / 10) + C,
    errors=lambda x, Aerr, Cerr, popt: sqrt(
        (Aerr * exp(x / popt[1])) ** 2 + Cerr**2
    ),
)
y3 = d.extrapolate(extra_x, overlap=80, kind="cubic")
y2 = d.extrapolate(extra_x, overlap=3.0, kind="quadratic")
y1 = d.extrapolate(extra_x, overlap=1.0, kind="linear")

d.plot(fmt="k.", capsize=2.0)
d.title = "Extrapolation Demo"

errorfill(
    extra_x, y4[:, 0], color="orange", yerr=y4[:, 1], label="Arg. Function"
)

errorfill(
    extra_x, y3[:, 0], color="red", yerr=y3[:, 1], label="Cubic Extrapolation"
)

errorfill(
    extra_x,
    y2[:, 0],
    color="green",
    yerr=y2[:, 1],
    label="Quadratic Extrapolation",
)
errorfill(
    extra_x, y1[:, 0], color="blue", yerr=y1[:, 1], label="Linear Extrapolation"
)
plt.legend(loc=2)



"""Sub-package of routines to load Data objects in different formats.

This sub-package is lazy-loaded but then pulls in all of the modules in order to register the routines.
"""

__all__ = ["generic", "hdf5", "instruments", "facilities", "rigs", "simulations", "attocube", "maximus", "zip"]

from . import generic, hdf5, instruments, facilities, rigs, simulations, attocube, maximus, zip



# -*- coding: utf-8 -*-
"""Functions for manipulating Kerr (or any other) images.

All of these functions are accessible through the :class:`ImageArray` attributes e.g.:

    k=ImageArray('myfile'); k.level_image().

The first 'im' function argument is automatically added in this case.

If you want to add new functions that's great. There's a few important points:

    * Please make sure they take an image as the first argument

    * Don't give them the same name as functions from the numpy library or
          skimage library if you don't want to override them.

    * The function should not change the shape of the array. Please use crop_image
          before doing the function if you want to do that.

    * After that you're free to treat im as a ImageArray
          or numpy array, it should all behave the same.
"""
__all__ = [
    "adjust_contrast",
    "align",
    "convert",
    "correct_drift",
    "subtract_image",
    "fft",
    "filter_image",
    "gridimage",
    "hist",
    "imshow",
    "level_image",
    "normalise",
    "profile_line",
    "quantize",
    "radial_coordinates",
    "radial_profile",
    "remove_outliers",
    "rotate",
    "translate",
    "translate_limits",
    "plot_histogram",
    "threshold_minmax",
    "do_nothing",
    "denoise",
]
import warnings
import os
import io

import numpy as np
from scipy.interpolate import griddata
from scipy.ndimage import gaussian_filter
from scipy import signal
from matplotlib.colors import to_rgba
import matplotlib.cm as cm
from matplotlib import pyplot as plt
from skimage import measure, transform, filters
from PIL import Image, PngImagePlugin

from Stoner.tools import isTuple, isiterable, make_Data

# from .core import ImageArray
from ..core.base import metadataObject
from .util import sign_loss, _dtype2, _supported_types, prec_loss, dtype_range, _dtype, _scale as im_scale
from ..tools.decorators import changes_size, keep_return_type
from .widgets import LineSelect
from ..compat import string_types, get_filedialog  # Some things to help with Python2 and Python3 compatibility
from ..plot.utils import auto_fit_fontsize

try:
    from PyQt5.QtGui import QImage
    from PyQt5.QtWidgets import QApplication
except ImportError:
    QImage = None
    QApplication = None

try:  # Make OpenCV an optional import
    import cv2
except ImportError:
    cv2 = None

try:  # Make imreg_dfft2 optional
    import imreg_dft
except ImportError:
    imreg_dft = None

try:  # image_registration module
    from image_registration import chi2_shift
except ImportError:
    chi2_shift = None

IMAGE_FILES = [("Tiff File", "*.tif;*.tiff"), ("PNG files", "*.png", "Numpy Files", "*.npy")]


def _scale(coord, scale=1.0, to_pixel=True):
    """Convert pixel coordinates to scaled coordinates or visa versa.

    Args:
        coord(int,float or iterable): Coordinates to be scaled

    Keyword Arguments:
        scale(float): Microns per Pixel scale of image
        to_pixel(bool): Force the conversion to be to pixels

    Returns:
        scaled coordinates.
    """
    if isinstance(coord, int):
        if not to_pixel:
            coord = float(coord) * scale
    elif isinstance(coord, float):
        if to_pixel:
            coord = int(round(coord / scale))
    elif isiterable(coord):
        coord = tuple([_scale(c, scale, to_pixel) for c in coord])
    else:
        raise ValueError("coord should be an integer or a float or an iterable of integers and floats")
    return coord


def adjust_contrast(im, lims=(0.1, 0.9), percent=True):
    """Rescale the intensity of the image.

    Mostly a call through to skimage.exposure.rescale_intensity. The absolute limits of contrast are
    added to the metadata as 'adjust_contrast'

    Parameters
    ----------
    lims: 2-tuple
        limits of rescaling the intensity
    percent: bool
        if True then lims are the give the percentile of the image intensity
        histogram, otherwise lims are absolute

    Returns
    -------
    image: ImageArray
        rescaled image
    """
    if percent:
        vmin, vmax = np.percentile(im, np.array(lims) * 100)
    else:
        vmin, vmax = lims[0], lims[1]
    im.metadata["adjust_contrast"] = (vmin, vmax)
    im = im.rescale_intensity(in_range=(vmin, vmax))  # clip the intensity
    return im


def _align_scharr(im, ref, **kargs):
    """Return the translation vector to shirt im to align to ref using the Scharr method."""
    scale = np.ceil(np.max(im.shape) / 500.0)
    ref1 = filters.edges.scharr(gaussian_filter(ref, sigma=scale, mode="wrap"))
    im1 = filters.edges.scharr(gaussian_filter(im, sigma=scale, mode="wrap"))
    return _align_imreg_dft(im1, ref1, **kargs)


def _align_chi2_shift(im, ref, **kargs):
    """Return the translation vector to shirt im to align to ref using the chi^2 shift method."""
    results = np.array(chi2_shift(ref, im, **kargs))
    return -results[1::-1], {}


def _align_imreg_dft(im, ref, **kargs):
    """Return the translation vector to shirt im to align to ref using the imreg_dft method."""
    constraints = kargs.pop("constraints", {"angle": [0.0, 0.0], "scale": [1.0, 0.0]})
    with warnings.catch_warnings():  # This causes a warning due to the masking
        warnings.simplefilter("ignore")
        result = imreg_dft.similarity(ref, im, constraints=constraints)
    tvec = result["tvec"]
    return np.array(tvec), result


def _align_cv2(im, ref, **kargs):  # pylint: disable=unused-argument
    """Return the translation vector to shirt im to align to ref using the cv2 method."""
    im1_gray = ref.convert("uint8", force_copy=True)
    im2_gray = im.convert("uint8", force_copy=True)

    # Define the motion model
    warp_mode = cv2.MOTION_TRANSLATION
    warp_matrix = np.eye(2, 3, dtype=np.float32)

    # Specify the number of iterations.
    number_of_iterations = 5000

    # Specify the threshold of the increment
    # in the correlation coefficient between two iterations
    termination_eps = 1e-10

    # Define termination criteria
    criteria = (cv2.TERM_CRITERIA_EPS | cv2.TERM_CRITERIA_COUNT, number_of_iterations, termination_eps)

    # Run the ECC algorithm. The results are stored in warp_matrix.
    (_, warp_matrix) = cv2.findTransformECC(
        im1_gray, im2_gray, warp_matrix, warp_mode, criteria, inputMask=None, gaussFiltSize=1
    )

    return -warp_matrix[1::-1, 2], {}


def align(im, ref, method="scharr", **kargs):
    """Use one of a variety of algroithms to align two images.

    Args:
        im (ndarray) image to align
        ref (ndarray) reference array

    Keyword Args:
        method (str or None):
            If given specifies which module to try and use.
            Options: 'scharr', 'chi2_shift', 'imreg_dft', 'cv2'
        _box (integer, float, tuple of images or floats):
            Used with ImageArray.crop to select a subset of the image to use for the aligning process.
        scale (int):
            Rescale the image and reference image by constant factor before finding the translation vector.
        prefilter (callable):
            A method to apply to the image before carrying out the translation to the align to the reference.
        **kargs (various): All other keyword arguments are passed to the specific algorithm.


    Returns
        (ImageArray or ndarray) aligned image

    Notes:
        Currently three algorithms are supported:
            - image_registration module's chi^2 shift: This uses a dft with an automatic
              up-sampling of the fourier transform for sub-pixel alignment. The metadata
              key *chi2_shift* contains the translation vector and errors.
            - imreg_dft module's similarity function. This implements a full scale, rotation, translation
              algorithm (by default constrained for just translation). It's unclear how much sub-pixel translation
              is accommodated.
            - cv2 module based affine transform on a gray scale image.
              from: http://www.learnopencv.com/image-alignment-ecc-in-opencv-c-python/
    """
    # To be consistent with x-y coordinate systems
    align_methods = {
        "scharr": (_align_scharr, imreg_dft),
        "chi2_shift": (_align_chi2_shift, chi2_shift),
        "imreg_dft": (_align_imreg_dft, imreg_dft),
        "cv2": (_align_cv2, cv2),
    }
    cls = type(im)
    for meth in list(align_methods.keys()):
        mod = align_methods[meth][1]
        if mod is None:
            del align_methods[meth]
    method = method.lower()
    if not len(align_methods):
        raise ImportError("align requires one of imreg_dft, chi2_shift or cv2 modules to be available.")
    if method not in align_methods:
        raise ValueError(f"{method} is not available either because it is not recognised or there is a missing module")

    if "box" in kargs or "_box" in kargs:
        box = kargs.pop("box", kargs.pop("_box"))
        if not isiterable(box):
            box = [box]
        working = im.crop(*box, copy=True)
        if ref.shape != working.shape:
            ref = ref.view(cls).crop(*box, copy=True)
    else:
        working = im

    scale = kargs.pop("scale", None)
    mode = kargs.pop("mode", "mirror")
    cval = kargs.pop("cval", im.mean())
    if mode == "mean":
        mode = "constant"
        cval = im.mean()

    if scale:
        working = working.rescale(scale, order=3)
        ref = transform.rescale(ref, scale, order=3)

    prefilter = kargs.pop("prefilter", True)
    try:
        tvec, data = align_methods[method][0](working, ref, **kargs)
    except (ValueError, TypeError):
        tvec = (0, 0)
        data = im

    data.pop("timg", None)

    if scale:
        tvec /= scale
    new_im = im.shift((tvec[1], tvec[0]), prefilter=prefilter, mode=mode, cval=cval)
    new_im.metadata.update(data)
    new_im["tvec"] = tuple(tvec)
    new_im["translation_limits"] = new_im.translate_limits("tvec")
    return new_im


def convert(image, dtype, force_copy=False, uniform=False, normalise=True):
    """Convert an image to the requested data-type.

    Warnings are issued in case of precision loss, or when negative values
    are clipped during conversion to unsigned integer types (sign loss).

    Floating point values are expected to be normalized and will be clipped
    to the range [0.0, 1.0] or [-1.0, 1.0] when converting to unsigned or
    signed integers respectively.

    Numbers are not shifted to the negative side when converting from
    unsigned to signed integer types. Negative values will be clipped when
    converting to unsigned integers.

    Parameters
    ----------
    image : ndarray
        Input image.
    dtype : dtype
        Target data-type.
    force_copy : bool
        Force a copy of the data, irrespective of its current dtype.
    uniform : bool
        Uniformly quantize the floating point range to the integer range.
        By default (uniform=False) floating point values are scaled and
        rounded to the nearest integers, which minimizes back and forth
        conversion errors.
    normalise : bool
        When converting from int types to float normalise the resulting array
        by the maximum allowed value of the int type.

    References
    ----------
    (1) DirectX data conversion rules.
        http://msdn.microsoft.com/en-us/library/windows/desktop/dd607323%28v=vs.85%29.aspx
    (2) Data Conversions.
        In "OpenGL ES 2.0 Specification v2.0.25", pp 7-8. Khronos Group, 2010.
    (3) Proper treatment of pixels as integers. A.W. Path.
        In "Graphics Gems I", pp 249-256. Morgan Kaufmann, 1990.
    (4) Dirty Pixels. J. Blinn.
        In "Jim Blinn's corner: Dirty Pixels", pp 47-57. Morgan Kaufmann, 1998.

    """
    image = np.asarray(image)
    dtypeobj = np.dtype(dtype)
    dtypeobj_in = image.dtype
    dtype = dtypeobj.type
    dtype_in = dtypeobj_in.type

    if dtype_in == dtype:
        if force_copy:
            image = image.clone
        return image

    if not (dtype_in in _supported_types and dtype in _supported_types):
        raise ValueError(f"can not convert {dtype_in} to {dtype}.")

    kind = dtypeobj.kind
    kind_in = dtypeobj_in.kind
    itemsize = dtypeobj.itemsize
    itemsize_in = dtypeobj_in.itemsize

    if kind == "b":
        # to binary image
        if kind_in in "fi":
            sign_loss(dtype_in, dtypeobj)
        prec_loss(dtypeobj_in, dtypeobj)
        return image > dtype_in(dtype_range[dtype_in][1] / 2)

    if kind_in == "b":
        # from binary image, to float and to integer
        result = np.where(~image, *dtype_range[dtype])
        return result

    if kind in "ui":
        imin = np.iinfo(dtype).min
        imax = np.iinfo(dtype).max
    if kind_in in "ui":
        imin_in = np.iinfo(dtype_in).min
        imax_in = np.iinfo(dtype_in).max

    if kind_in == "f":
        if np.min(image) < -1.0 or np.max(image) > 1.0:
            raise ValueError("Images of type float must be between -1 and 1.")
        if kind == "f":
            # floating point -> floating point
            if itemsize_in > itemsize:
                prec_loss(dtypeobj_in, dtypeobj)
            return image.astype(dtype)

        # floating point -> integer
        prec_loss(dtypeobj_in, dtypeobj)
        # use float type that can represent output integer type
        image = np.array(image, _dtype(itemsize, dtype_in, np.float32, np.float64))
        if not uniform:
            if kind == "u":
                image *= imax
            elif itemsize_in <= itemsize and itemsize == 8:  # f64->int64 needs care to avoid overruns!
                image *= 2**54  # float64 has 52bits of mantissa, so this will avoid precission loss for a +/-1 range
                np.rint(image, out=image)
                image = image.astype(dtype)
                np.clip(image, -(2**54), 2**54 - 1, out=image)
                image *= 512
            else:
                image *= imax - imin
                image -= 1.0
                image /= 2.0
                np.rint(image, out=image)
                np.clip(image, imin, imax, out=image)
        elif kind == "u":
            image *= imax + 1
            np.clip(image, 0, imax, out=image)
        else:
            image *= (imax - imin + 1.0) / 2.0
            np.floor(image, out=image)
            np.clip(image, imin, imax, out=image)
        return image.astype(dtype)

    if kind == "f":
        # integer -> floating point
        if itemsize_in >= itemsize:
            prec_loss(dtypeobj_in, dtypeobj)
        # use float type that can exactly represent input integers
        image = np.array(image, _dtype(itemsize_in, dtype, np.float32, np.float64))
        if normalise:  # normalise floats by maximum value of int type
            if kind_in == "u":
                image /= imax_in
                # DirectX uses this conversion also for signed ints
                # if imin_in:
                #    np.maximum(image, -1.0, out=image)
            else:
                image *= 2.0
                image += 1.0
                image /= imax_in - imin_in
        return image.astype(dtype)

    if kind_in == "u":
        if kind == "i":
            # unsigned integer -> signed integer
            image = im_scale(image, 8 * itemsize_in, 8 * itemsize - 1, dtypeobj_in, dtypeobj)
            return image.view(dtype)
        # unsigned integer -> unsigned integer
        return im_scale(image, 8 * itemsize_in, 8 * itemsize, dtypeobj_in, dtypeobj)

    if kind == "u":
        # signed integer -> unsigned integer
        sign_loss(dtype_in, dtypeobj)
        image = im_scale(image, 8 * itemsize_in - 1, 8 * itemsize, dtypeobj_in, dtypeobj)
        result = np.empty(image.shape, dtype)
        np.maximum(image, 0, out=result, dtype=image.dtype, casting="unsafe")
        return result

    # signed integer -> signed integer
    if itemsize_in > itemsize:
        return im_scale(image, 8 * itemsize_in - 1, 8 * itemsize - 1, dtypeobj_in, dtypeobj)
    image = image.astype(_dtype2("i", itemsize * 8))
    image -= imin_in
    image = im_scale(image, 8 * itemsize_in, 8 * itemsize, dtypeobj_in, dtypeobj, copy=False)
    image += imin
    return image.astype(dtype)


def correct_drift(im, ref, **kargs):
    """Align images to correct for image drift.

    Args:
        ref (ImageArray): Reference image with assumed zero drift

    Keyword Arguments:
        threshold (float): threshold for detecting imperfections in images
            (see skimage.feature.corner_fast for details)
        upsample_factor (float): the resolution for the shift 1/upsample_factor pixels registered.
            see skimage.feature.register_translation for more details
        box (sequence of 4 ints): defines a region of the image to use for identifyign the drift
            defaults to the whol image. Use this to avoid drift calculations being confused by
            the scale bar/annotation region.
        do_shift (bool): Shift the image, or just calculate the drift and store in metadata (default True, shit)

    Returns:
        A shifted image with the image shift added to the metadata as 'correct drift'.

    Detects common features on the images and tracks them moving.
    Adds 'drift_shift' to the metadata as the (x,y) vector that translated the
    image back to it's origin.
    """
    do_shift = kargs.pop("do_shift", True)
    kargs["scale"] = kargs.pop("upscale", kargs.get("scale", 5.0))
    kargs.setdefault("meothd", "scharr")

    ret = align(im, ref, **kargs)
    if do_shift:
        im = ret
    im["correct_drift"] = -np.array(ret["tvec"])[::-1]

    return im


def subtract_image(im, background, contrast=16, clip=True, offset=0.5):
    """Subtract a background image from the ImageArray.

    Multiply the contrast by the contrast parameter.
    If clip is on then clip the intensity after for the maximum allowed data range.
    """
    im = im.asfloat(normalise=False, clip_negative=False)
    im = contrast * (im - background) + offset
    if clip:
        im = im.clip_intensity()
    return im


def fft(im, shift=True, phase=False, remove_dc=False, gaussian=None, window=None):
    """Perform a 2d fft of the image and shift the result to get zero frequency in the centre.

    Keyword Args:
        shift (bool):
            Shift the fft so that zero order is in the centre of the image. Default True
        phase (bool, None):
            If true, return the phase angle rather than the magnitude if False. If None, return the raw fft.
            Default False - return magnitude of fft.
        remove_dc(bool):
            Replace the points around the dc offset with the mean of the fft to avoid dc offset artefacts.
            Default False
        gaussian (None or float):
            Apply a gaussian blur to the fft where this parameter is the width of the blue in px. Default None for off.
        window (None or str):
            If not None (default) the image is multiplied by the given window function before the fft is calculated.
            This avpoids leaking some signal into the higher frequency bands due to discontinuities at the image edges.

    Return:
        fft of the image, preserving metadata.
    """
    if window:
        window = filters.window(window, im.shape)
        im = im.clone * window
    r = np.fft.fft2(im)

    if remove_dc:
        fill = r.mean()
        r[0, 0] = fill
        r[-1, 0] = fill
        r[-1, -1] = fill
        r[0, -1] = fill
    if shift:
        r = np.fft.fftshift(r)
    if phase is None:
        pass
    if not phase:
        r = np.abs(r)
    else:
        r = np.angle(r)
    r = r.view(type(im))
    if isinstance(gaussian, (float, int)):
        r.gaussian(gaussian)

    r.metadata.update(im.metadata)
    return r


def filter_image(im, sigma=2):
    """Alias for skimage.filters.gaussian."""
    return im.gaussian(sigma=sigma)


def gridimage(im, points=None, xi=None, method="linear", fill_value=None, rescale=False):
    """Use :py:func:`scipy.interpolate.griddata` to shift the image to a regular grid of coordinates.

    Args:
        points (tuple of (x-co-ords,yco-ordsa)):
            The actual sampled data coordinates
        xi (tupe of (2D array,2D array)):
            The regular grid coordinates (as generated by e.g. :py:func:`np.meshgrid`)

    Keyword Arguments:
        method ("linear","cubic","nearest"):
            how to interpolate, default is linear
        fill_value (folat, Callable, None):
            What to put when the coordinates go out of range (default is None). May be a callable
            in which case the initial image is presented as the only argument. If None, use the mean value.
        rescale (bool):
            If the x and y coordinates are very different in scale, set this to True.

    Returns:
        A copy of the modified image. The image data is interpolated and metadata kets "actual_x","actual_y","sample_
        x","samp[le_y" are set to give coordinates of new grid.

    Notes:
        If points and or xi are missed out then we try to construct them from the metadata. For points, the metadata
        keys "actual-x" and "actual_y" are looked for and for xi, the metadata keys "sample_x" and "sample_y" are
        used. These are set, for example, by the :py:class:`Stoner.HDF5.SXTMImage` loader if the interformeter stage
        data was found in the file.

        The metadata used in this case is then adjusted as well to ensure that repeated application of this method
        doesn't change the image after it has been corrected once.
    """
    if points is None:
        points = np.column_stack((im["actual_x"].ravel(), im["actual_y"].ravel()))
    if xi is None:
        xi = xi = (im["sample_x"], im["sample_y"])

    if fill_value is None:
        fill_value = np.mean

    if callable(fill_value):
        fill_value = fill_value(im)

    im2 = griddata(points, im.ravel(), xi, method, fill_value, rescale)
    im2 = type(im)(im2)
    im2.metadata = im.metadata
    im2.metadata["actual_x"] = xi[0]
    im2.metadata["actual_y"] = xi[1]
    return im2


def hist(im, *args, **kargs):
    """Pass through to :py:func:`matplotlib.pyplot.hist` function."""
    if isinstance(im, np.ma.MaskedArray):
        im_data = im[~im.mask]
    else:
        im_data = im.ravel()
    counts, edges = np.histogram(im_data, *args, **kargs)
    centres = (edges[1:] + edges[:-1]) / 2
    new = make_Data(np.column_stack((centres, counts)))
    new.column_headers = ["Intensity", "Frequency"]
    new.setas = "xy"
    return new


def imshow(im, **kwargs):
    """Quickly plot of image.

    Keyword Arguments:
        figure (int, str or matplotlib.figure):
            if int then use figure number given, if figure is 'new' then create a new figure, if None then use
            whatever default figure is available
        show_axis (bool):
            If True, show the axis otherwise don't (default)'
        title (str,None,False):
            Title for plot - defaults to False (no title). None will take the title from the filename if present
        title_args (dict):
            Arguments to pass to the title function to control formatting.
        cmap (str,matplotlib.cmap):
            Colour scheme for plot, defaults to gray

    Any masked areas are set to NaN which stops them being plotted at all.
    """
    figure = kwargs.pop("figure", "new")
    ax = kwargs.pop("ax", None)
    # Get a title - from keyword argument, from title attr or filename attr
    title = os.path.split(getattr(im, "title", getattr(im, "filename", "")))[-1]
    title = kwargs.pop("title", title)
    title_args = kwargs.pop("title_args", {})
    cmap = kwargs.pop("cmap", "gray")
    mask_alpha = kwargs.pop("mask_alpha", getattr(im, "_mask_alpha", 0.5))
    mask_col = kwargs.pop("mask_color", getattr(im, "_mask_color", "red"))
    if isinstance(cmap, string_types):
        cmap = getattr(cm, cmap)
    im_data = im
    if figure is not None and isinstance(figure, int):
        fig = plt.figure(figure)
    elif figure is not None and figure == "new":
        fig = plt.figure()
    elif figure is not None:  # matplotlib.figure instance
        fig = plt.figure(figure.number)
    else:
        fig = plt.figure()
    if ax is not None:
        plt.sca(ax)
    else:
        ax = plt.gca()
    ax.imshow(im_data.view(np.ndarray), cmap=cmap, **kwargs)
    if np.ma.is_masked(im):
        mask_col = list(to_rgba(mask_col))
        mask_col[-1] = mask_alpha
        mask_data = np.zeros(im.shape + (4,))
        mask_data[im.mask] = mask_col
        ax.imshow(mask_data)

    if title is None:
        if "filename" in im.metadata.keys():
            title = os.path.split(im["filename"])[1]
        elif hasattr(im, "filename"):
            title = os.path.split(im.filename)[1]
        else:
            title = " "
    elif isinstance(title, bool) and not title:
        title = ""

    txt = ax.set_title(title, **title_args)
    if not title_args.get("fontdict", {}).get("fontsize", False):
        gs = ax.get_gridspec()
        width = 0.9 / gs.ncols
        height = 0.09 / gs.nrows
        auto_fit_fontsize(txt, width, height)
    ax.axis("on" if kwargs.get("show_axis", False) else "off")

    if QImage is None:  # No Qt5
        return fig

    def add_figure_to_clipboard(event):
        if event.key == "ctrl+c":
            with io.BytesIO() as buffer:
                fig.savefig(buffer)
                QApplication.clipboard().setImage(QImage.fromData(buffer.getvalue()))

    fig.canvas.mpl_connect("key_press_event", add_figure_to_clipboard)
    return fig


def level_image(im, poly_vert=1, poly_horiz=1, box=None, poly=None, mode="clip"):
    """Subtract a polynomial background from image.

    Keyword Arguments:
        poly_vert (int): fit a polynomial in the vertical direction for the image of order
            given. If 0 do not fit or subtract in the vertical direction
        poly_horiz (int): fit a polynomial of order poly_horiz to the image. If 0 given
            do not subtract
        box (array, list or tuple of int): [xmin,xmax,ymin,ymax] define region for fitting. IF None use entire
            image
        poly (list or None): [pvert, phoriz] pvert and phoriz are arrays of polynomial coefficients
            (highest power first) to subtract in the horizontal and vertical
            directions. If None function defaults to fitting its own polynomial.
        mode (str): Either 'clip' or 'norm' - specifies how to handle intensitry values that end up being outside
            of the accepted range for the image.

    Returns:
        A new copy of the processed images.

    Fit and subtract a background to the image. Fits a polynomial of order
    given in the horizontal and vertical directions and subtracts. If box
    is defined then level the *entire* image according to the
    gradient within the box. The polynomial subtracted is added to the
    metadata as 'poly_vert_subtract' and 'poly_horiz_subtract'
    """
    if box is None:
        box = im.max_box
    cim = im.crop(box=box)
    (vertl, horizl) = cim.shape
    p_horiz = 0
    p_vert = 0
    if poly_horiz > 0:
        comp_vert = np.average(cim, axis=0)  # average (compress) the vertical values
        if poly is not None:
            p_horiz = poly[0]
        else:
            p_horiz = np.polyfit(np.arange(horizl), comp_vert, poly_horiz)  # fit to the horizontal
            av = np.average(comp_vert)  # get the average pixel height
            p_horiz[-1] = p_horiz[-1] - av  # maintain the average image height
        horizcoord = np.indices(im.shape)[1]  # now apply level to whole image
        for i, c in enumerate(p_horiz):
            im = im - c * horizcoord ** (len(p_horiz) - i - 1)
    if poly_vert > 0:
        comp_horiz = np.average(cim, axis=1)  # average the horizontal values
        if poly is not None:
            p_vert = poly[1]
        else:
            p_vert = np.polyfit(np.arange(vertl), comp_horiz, poly_vert)
            av = np.average(comp_horiz)
            p_vert[-1] = p_vert[-1] - av  # maintain the average image height
        vertcoord = np.indices(im.shape)[0]
        for i, c in enumerate(p_vert):
            im = im - c * vertcoord ** (len(p_vert) - i - 1)
    im.metadata["poly_sub"] = (p_horiz, p_vert)
    if mode == "clip":
        im = im.clip_intensity()  # saturate any pixels outside allowed range
    elif mode == "norm":
        im = im.normalise()
    return im


def normalise(im, scale=None, sample=False, limits=(0.0, 1.0), scale_masked=False):
    """Norm alise the data to a fixed scale.

    Keyword Arguments:
        scale (2-tuple):
            The range to scale the image to, defaults to -1 to 1.
        saple (box):
            Only use a section of the input image to calculate the new scale over. Default is False - whole image
        limits (low,high):
            Take the input range from the *high* and *low* fraction of the input when sorted.
        scale_masked (bool):
            If True then the masked region is also scaled, otherwise any masked region is ignored. Default, False.


    Returns:
        A scaled version of the data. The ndarray min and max methods are used to allow masked images
        to be operated on only on the unmasked areas.

    Notes:
        The *sample* keyword controls the area in which the range of input values is calculated, the actual scaling is
        done on the whole image.

        The *limits* parameter is used to set the input scale being normalised from - if an image has a few outliers
        then this setting can be used to clip the input range before normalising. The parameters in the limit are the
        values at the *low* and *high* fractions of the cumulative distribution functions.
    """
    mask = im.mask
    cls = type(im)
    im = im.astype(float)
    if scale is None:
        scale = (-1.0, 1.0)
    section = im[im._box(sample)]

    section = section[~section.mask]
    if limits != (0.0, 1.0):
        low, high = limits
        low = np.sort(section.ravel())[int(low * section.size)]
        high = np.sort(section.ravel())[int(high * section.size)]
        im.clip_intensity(limits=(low, high))
    else:
        high = section.max()
        low = section.min()

    if not isTuple(scale, float, float, strict=False):
        raise ValueError("scale should be a 2-tuple of floats.")
    scaled = (im.data - low) / (high - low)
    delta = scale[1] - scale[0]
    offset = scale[0]
    scaled = scaled * delta + offset
    if not scale_masked:
        im = np.where(im.mask, im, scaled).view(cls)
    else:
        im = scaled.view(cls)
    im.mask = mask
    return im


def clip_neg(im):
    """Clip negative pixels to 0.

    Most useful for float where pixels above 1 are reduced to 1.0 and -ve pixels
    are changed to 0.
    """
    im[im < 0] = 0
    return im


def profile_line(img, src=None, dst=None, linewidth=1, order=1, mode="constant", cval=0.0, constrain=True, **kargs):
    """Wrap sckit-image method of the same name to get a line_profile.

    Parameters:
        img(ImageArray):
            Image data to take line section of

    Keyword Parameters:
        src, dst (2-tuple of int or float):
            start and end of line profile. If the coordinates
            are given as integers then they are assumed to be pxiel coordinates, floats are
            assumed to be real-space coordinates using the embedded metadata.
        linewidth (int):
            the wideth of the profile to be taken.
        order (int 1-3):
            Order of interpolation used to find image data when not aligned to a point
        mode (str):
            How to handle data outside of the image.
        cval (float):
            The constant value to assume for data outside of the image is mode is "constant"
        constrain (bool):
            Ensure the src and dst are within the image (default True).
        no_scale (bool):
            Do not attempt to scale values by the image scale and work in pixels throughout. (default False)


    Returns:
        A :py:class:`Stoner.Data` object containing the line profile data and the metadata from the image.
    """
    scale = 1.0 if kargs.get("no_scale", False) else img.get("MicronsPerPixel", 1.0)
    r, c = img.shape
    fast_mode = False
    if src is None and dst is None:
        if "x" in kargs:
            src = (0, kargs["x"])
            dst = (r, kargs["x"])
            fast_mode = kargs.get("no_scale", False)
        if "y" in kargs:
            src = (kargs["y"], 0)
            dst = (kargs["y"], c)
            fast_mode = kargs.get("no_scale", False)
        if src is None and dst is None:
            src, dst = LineSelect()(img)
    if isinstance(src, float):
        src = (src, src)
    if isinstance(dst, float):
        dst = (dst, dst)
    dst = _scale(dst, scale)
    src = _scale(src, scale)
    if not isTuple(src, int, int):
        raise ValueError("src coordinates are not a 2-tuple of ints.")
    if not isTuple(dst, int, int):
        raise ValueError("dst coordinates are not a 2-tuple of ints.")

    if constrain:
        fix = lambda x, mx: int(round(sorted([0, x, mx])[1]))
        r, c = img.shape
        src = list(src)
        src = (fix(src[0], r), fix(src[1], c))
        dst = (fix(dst[0], r), fix(dst[1], c))

    if fast_mode:
        if "x" in kargs:
            result = img[:, src[1]]
            points = np.row_stack((np.ones(img.shape[0]) * src[1], np.arange(img.shape[0])))
        else:
            result = img[src[0], :]
            points = np.row_stack((np.arange(img.shape[1]), np.ones(img.shape[1]) * src[0]))
    else:
        result = measure.profile_line(img, src, dst, linewidth, order, mode, cval)
        points = measure.profile._line_profile_coordinates(src, dst, linewidth)[:, :, 0]
    ret = make_Data()
    ret.data = points.T
    ret.setas = "xy"
    x, y = points
    x -= x[0]
    y -= y[0]
    ret &= np.sqrt(x**2 + y**2) * scale
    ret &= result
    ret.column_headers = ["X", "Y", "Distance", "Intensity"]
    ret.setas = "..xy"
    ret.metadata = img.metadata.copy()
    return ret


def radial_coordinates(im, centre=(None, None), pixel_size=(1, 1), angle=False):
    """Rerurn a map of the radial coordinates of an image from a given centre, with adjustments for pixel size.

    Keyword Arguments:
        centre (2-tuple):
            Coordinates of centre point in terms of the original pixels. Defaults to(None,None) for the middle of the
            image.
        pixel_size (2-tuple):
            The size of one pixel in (dx by dy) - defaults to 1,1
        angle (bool, None):
            Whether to return the angles (in radians, True), distances (False) o a complex number (None).

    Returns:
        An array of the same class as the input, but with values corresponding to the radial coordinates.
    """
    cx, cy = centre
    r, c = im.shape
    dx, dy = pixel_size
    cx = c / 2 if cx is None else cx
    cy = r / 2 if cy is None else cy
    x_r = dx * (np.linspace(0, c - 1, c) - cx)
    y_r = dy * (np.linspace(0, r - 1, r) - cy)
    Y, X = np.meshgrid(x_r, y_r)
    Z = -Y + (0 + 1j) * X
    if angle is None:
        pass
    elif not angle:
        Z = np.abs(Z)
    else:
        Z = np.angle(Z)
    Z = Z.view(type(im))
    Z.metadata = im.metadata
    return Z


def radial_profile(im, angle=None, r=None, centre=(None, None), pixel_size=(1, 1)):
    """Extract a radial  profile line from an image.

    Keyword Parameters:
        angle (float, tuple, None):
            Select the radial angle to include:
                - float selects a single angle
                - tuple selects a range of angles
                - None integrates over all angles
        r (array, None):
            Edges of the bins in the radual direction - will return r.size-1 points. Default is None which uses the
            minimum r value found on the edges of the image.
        centre (2-tuple):
            Coordinates of centre point in terms of the original pixels. Defaults to(None,None) for the middle of the
            image.
        pixel_size (2-tuple):
            The size of one pixel in (dx by dy) - defaults to 1,1

    Returns:
        (Data):
            A py:class:`Stoner.Data` object with a column for r and columns for mean, std, and number of pixels.
    """
    coords = im.radial_coordinates(centre=centre, pixel_size=pixel_size, angle=None)
    if r is None:  # Identify the minimum edge value
        edges = np.append(coords[:, 0], coords[-1, :])
        edges = np.append(edges, coords[:, -1])
        edges = np.append(edges, coords[0, :])
        r_limit = np.abs(edges).min()
        r = np.linspace(0, np.ceil(r_limit), int(np.ceil(r_limit) + 1))
    r_l = r[:-1]
    r_h = r[1:]
    r_m = (r_l + r_h) / 2
    if angle is None:
        angle_select = np.ones_like(coords).astype(bool)
    elif isinstance(angle, tuple):
        angle_low = np.angle(coords) >= angle[0]
        angle_high = np.angle(coords) <= angle[1]
        angle_select = np.logical_and(angle_low, angle_high)
    elif isinstance(angle, (int, float)):
        angle_select = np.isclose(np.angle(coords), angle)
    else:
        raise TypeError(f"angle should be a float, tuple of two floats or None not a {type(angle)}")
    ret = make_Data()
    for low, high, mid in zip(r_l, r_h, r_m):
        r_select = np.logical_and(np.abs(coords) >= low, np.abs(coords) < high)
        data = im[np.logical_and(r_select, angle_select)]
        if data.size == 0:
            continue
        if data.size == 1:
            avg = data[0]
            std = np.nan
        else:
            avg = data.mean()
            std = data.std()
        num = data.size
        ret += np.array([low, mid, high, avg, std, num])
    ret.column_headers = ["Low_r", "r", "high_r", "mean", "std", "number"]
    ret.setas = ".x.ye."
    ret.metadata = im.metadata
    ret.filename = im.filename[:-4] + "_profile.txt"
    return ret


def quantize(im, output, levels=None):
    """Quantise the image data into fixed levels given by a mapping.

    Args:
        output (list,array,tuple): Output levels to return.

    Keyword Arguments:
        levels (list, array or None): The input band markers. If None is constructed from the data.

    The number of levels should be one less than the number of output levels given.

    Notes:
        The routine will ignore all masked pixels and will preserve the mask.
    """
    section = im[~im.mask]
    mask = im.mask
    lmin, lmax = section.min(), section.max()  # Dudge to ensure that the bottom and top elements are included.
    delta = (lmax - lmin) / 100

    if levels is None:
        levels = np.linspace(lmin - delta, lmax + delta, len(output) + 1)
    elif len(levels) == len(output) + 1:
        pass
    elif len(levels) == len(output) - 1:
        lvl = np.zeros(len(output) + 1)
        lvl[1:-1] = levels
        lvl[0] = section.min() - delta
        lvl[-1] = section.max() + delta
        levels = lvl
    else:
        raise RuntimeError(f"{len(output)} output levels and {len(levels)} input levels")

    ret = im.clone
    ret.mask = False
    for lvl, lvh, val in zip(levels[:-1], levels[1:], output):
        select = np.logical_and(np.less_equal(im, lvh), np.greater(im, lvl))
        ret[select] = val
    ret[mask] = im.view(np.ndarray)[mask]  # Put back the original image values where they are masked.
    ret.mask = mask  # restore mask
    return ret


def remove_outliers(im, percentiles=(0.01, 0.99), replace=None):
    """Find values of the data that are beyond a percentile of the overall distribution and replace them.

    Keyword Parameters:
        percentile (2 tuple):
            Fraction percentiles to consider to be outliers (default is (0.01,0.99) for 1% limits)
        replace (2 tuple or None):
            Values to set outliers to. If None, then the pixel values at the percentile limits are used.

    Returns:
        (ndarray):
            The modified array.

    Use this method if you have an image with a small number of pixels with extreme values that are
    out of range.
    """
    from scipy.interpolate import interp1d

    cdf, bins = im.cumulative_distribution(nbins=min(1000, im.count()))
    cdf = interp1d(cdf, bins, kind="linear")
    limits = [cdf(x) for x in percentiles]
    if replace is None:
        replace = limits
    im[im < limits[0]] = replace[0]
    im[im > limits[1]] = replace[1]
    return im


def rotate(im, angle, resize=False, center=None, order=1, mode="constant", cval=0, clip=True, preserve_range=False):
    """Rotate image by a certain angle around its center.

    Parameters:
        angle  (float):
            Rotation angle in **radians** in clockwise direction.

    Keyword Parameters:
        resize (bool):
            Determine whether the shape of the output image will be automatically
            calculated, so the complete rotated image exactly fits. Default is
            False.
        center (iterable of length 2):
            The rotation center. If ``center=None``, the image is rotated around
            its center, i.e. ``center=(cols / 2 - 0.5, rows / 2 - 0.5)``.  Please
            note that this parameter is (cols, rows), contrary to normal skimage
            ordering.
        order (int):
            The order of the spline interpolation, default is 1. The order has to
            be in the range 0-5. See `skimage.transform.warp` for detail.
        mode ({'constant', 'edge', 'symmetric', 'reflect', 'wrap'}):
            Points outside the boundaries of the input are filled according
            to the given mode.  Modes match the behaviour of `numpy.pad`.
        cval  (float):
            Used in conjunction with mode 'constant', the value outside
            the image boundaries.
        clip (bool):
            Whether to clip the output to the range of values of the input image.
            This is enabled by default, since higher order interpolation may
            produce values outside the given input range.
        preserve_range (bool):
            Whether to keep the original range of values. Otherwise, the input
            image is converted according to the conventions of `Stpomer.Image.ImageArray.as_float`.

    Returns:
        (ImageFile/ImageArray):
            Rotated image

    Notes:
        (pass through to the skimage.transform.warps.rotate function)
    """
    ang = np.rad2deg(-angle)
    data = transform.rotate(
        im,
        ang,
        resize=resize,
        center=center,
        order=order,
        mode=mode,
        cval=cval,
        clip=clip,
        preserve_range=preserve_range,
    )
    print(type(im), type(data))
    ret = data.view(type(im))
    try:
        ret.metadata = im.metadata
        ret.metadata["transform:rotation"] = angle
    except AttributeError:
        pass
    return ret


def sgolay2d(img, points=15, poly=1, derivative=None):
    """Implements a 2D Savitsky Golay Filter for a 2D array (e.g. image).

    Arguments:
        img (ImageArray or ImageFile):
            image to be filtered

    Keyword Arguments:
        points (int):
            The number of points in the window aperture. Must be an odd number. (default 15)
        poly (int):
            Degree of polynomial to use in the filter. (default 1)
        derivative (str or None):
            Type of defivative to calculate. Can be:
                None - smooth only (default)
                "x","y" - calculate dIntentity/dx or dIntensity/dy
                "both" - calculate the full derivative and return magnitud and angle.

    ReturnsL
        (imageArray or ImageFile):
            filtered image.

    Raises:
        ValueError if points, order or derivative are incorrect.

    Notes:
        Adapted from code on the scipy cookbook : https://scipy-cookbook.readthedocs.io/items/SavitzkyGolay.html
    """
    # number of terms in the polynomial expression
    n_terms = (poly + 1) * (poly + 2) / 2.0

    if points % 2 == 0:
        raise ValueError("window_size must be odd")

    if points**2 < n_terms:
        raise ValueError("order is too high for the window size")

    half_size = points // 2

    # exponents of the polynomial.
    # p(x,y) = a0 + a1*x + a2*y + a3*x^2 + a4*y^2 + a5*x*y + ...
    # this line gives a list of two item tuple. Each tuple contains
    # the exponents of the k-th term. First element of tuple is for x
    # second element for y.
    # Ex. exps = [(0,0), (1,0), (0,1), (2,0), (1,1), (0,2), ...]
    exps = [(k - n, n) for k in range(poly + 1) for n in range(k + 1)]

    # coordinates of points
    ind = np.arange(-half_size, half_size + 1, dtype=np.float64)
    dx = np.repeat(ind, points)
    dy = np.tile(ind, [points, 1]).reshape(points**2)

    # build matrix of system of equation
    A = np.empty((points**2, len(exps)))
    for i, exp in enumerate(exps):
        A[:, i] = (dx ** exp[0]) * (dy ** exp[1])

    # pad input array with appropriate values at the four borders
    new_shape = img.shape[0] + 2 * half_size, img.shape[1] + 2 * half_size
    Z = np.zeros((new_shape))
    # top band
    band = img[0, :]
    Z[:half_size, half_size:-half_size] = band - np.abs(np.flipud(img[1 : half_size + 1, :]) - band)
    # bottom band
    band = img[-1, :]
    Z[-half_size:, half_size:-half_size] = band + np.abs(np.flipud(img[-half_size - 1 : -1, :]) - band)
    # left band
    band = np.tile(img[:, 0].reshape(-1, 1), [1, half_size])
    Z[half_size:-half_size, :half_size] = band - np.abs(np.fliplr(img[:, 1 : half_size + 1]) - band)
    # right band
    band = np.tile(img[:, -1].reshape(-1, 1), [1, half_size])
    Z[half_size:-half_size, -half_size:] = band + np.abs(np.fliplr(img[:, -half_size - 1 : -1]) - band)
    # central band
    Z[half_size:-half_size, half_size:-half_size] = img

    # top left corner
    band = img[0, 0]
    Z[:half_size, :half_size] = band - np.abs(np.flipud(np.fliplr(img[1 : half_size + 1, 1 : half_size + 1])) - band)
    # bottom right corner
    band = img[-1, -1]
    Z[-half_size:, -half_size:] = band + np.abs(
        np.flipud(np.fliplr(img[-half_size - 1 : -1, -half_size - 1 : -1])) - band
    )

    # top right corner
    band = Z[half_size, -half_size:]
    Z[:half_size, -half_size:] = band - np.abs(np.flipud(Z[half_size + 1 : 2 * half_size + 1, -half_size:]) - band)
    # bottom left corner
    band = Z[-half_size:, half_size].reshape(-1, 1)
    Z[-half_size:, :half_size] = band - np.abs(np.fliplr(Z[-half_size:, half_size + 1 : 2 * half_size + 1]) - band)

    # solve system and convolve
    if derivative is None:
        m = np.linalg.pinv(A)[0].reshape((points, -1))
        ret = signal.fftconvolve(Z, m, mode="valid").view(type(img))
    elif derivative == "x":
        c = np.linalg.pinv(A)[1].reshape((points, -1))
        ret = signal.fftconvolve(Z, -c, mode="valid").view(type(img))
    elif derivative == "y":
        r = np.linalg.pinv(A)[2].reshape((points, -1))
        ret = signal.fftconvolve(Z, -r, mode="valid").view(type(img))
    elif derivative == "both":
        c = np.linalg.pinv(A)[1].reshape((points, -1))
        r = np.linalg.pinv(A)[2].reshape((points, -1))
        ret = signal.fftconvolve(Z, -r, mode="valid"), signal.fftconvolve(Z, -c, mode="valid").view(type(img))
    else:
        raise ValueError(f"Unknown derivative mode {derivative}")
    ret.metadata.update(img.metadata)
    return ret


def span(im):
    """Return the minimum and maximum values in the image."""
    return np.min(im), np.max(im)


def translate(im, translation, add_metadata=False, order=3, mode="wrap", cval=None):
    """Translate the image.

    Areas lost by move are cropped, and areas gained are made black (0)
    The area not lost or cropped is added as a metadata parameter
    'translation_limits'

    Args:
        translate (2-tuple):
            translation (x,y)

    Keyword Arguments:
        add_metadata (bool):
            Record the shift in the image metadata order (int): Interpolation order (default, 3, bi-cubic)
        mode (str):
            How to handle points outside the original image. See :py:func:`skimage.transform.warp`.  Defaults to "wrap"
        cval (float):
            The value to fill with if *mode* is constant. If not specified or None, defaults to the mean pixcel value.

    Returns:
        im (ImageArray): translated image
    """
    translation = [-x for x in translation]
    trans = transform.SimilarityTransform(translation=translation)
    if cval is None:
        cval = im.mean()
    im = im.warp(trans, order=order, mode=mode, cval=cval)
    if add_metadata:
        im.metadata["translation"] = translation
        im.metadata["translation_limits"] = translate_limits(im, translation)
    return im


def translate_limits(im, translation, reverse=False):
    """Find the limits of an image after a translation.

    After using ImageArray.translate some areas will be black,
    this finds the max area that still has original pixels in

    Args:
        translation: 2-tuple
            the (x,y) translation applied to the image

    Keyword Args:
        reverse (bool):
            whether to reverse the translation vector (default False, no)

    Returns:
        limits: 4-tuple
            (xmin,xmax,ymin,ymax) the maximum coordinates of the image with original
            information
    """
    if isinstance(translation, string_types):
        translation = im[translation]

    translation = np.array(translation)
    if reverse:
        translation *= -1

    shape = im.shape

    xmin = max(0, translation[0])
    xmax = min(shape[0], shape[0] + translation[0])
    ymin = max(0, translation[1])
    ymax = min(shape[1], shape[1] + translation[1])

    return (xmin, xmax, ymin, ymax)


def plot_histogram(im, bins=256):
    """Plot the histogram and cumulative distribution for the image."""
    hist, bins = np.histogram(im, bins)
    cum, bins = im.cumulative_distribution(nbins=bins)
    cum = cum * np.max(hist) / np.max(cum)
    plt.figure()
    plt.plot(bins, hist, "k-")
    plt.plot(bins, cum, "r-")


def threshold_minmax(im, threshmin=0.1, threshmax=0.9):
    """Return a boolean array which is thresholded between threshmin and  threshmax.

    (ie True if value is between threshmin and threshmax)
    """
    im = im.convert(float)
    return np.logical_and(im > threshmin, im < threshmax)


def denoise(im, weight=0.1):
    """Rename the skimage restore function."""
    return im.denoise_tv_chambolle(weight=weight)


def do_nothing(self):
    """Nulop function for testing the integration into ImageArray."""
    return self


@changes_size
def crop(self, *args, **kargs):
    """Crop the image according to a box.

    Args:
        box(tuple) or 4 separate args or None:
            (xmin,xmax,ymin,ymax)
            If None image will be shown and user will be asked to select
            a box (bit experimental)

    Keyword Arguments:
        copy(bool):
            If True return a copy of ImageFile with the cropped image
    Returns:
        (ImageArray):
            view or copy of array asked for

    Notes:
        This is essentially like taking a view onto the array
        but uses image x,y coords (x,y --> col,row)
        Returns a view according to the coords given. If box is None it will
        allow the user to select a rectangle. If a tuple is given with None
        included then max extent is used for that coord (analogous to slice).
        If copy then return a copy of self with the cropped image.

        The box can be specified in a number of ways:
            -   (int):
                A border around all sides of the given number pixels is ignored.
            -   (float 0.0-1.0):
                A border of the given fraction of the images height and width is ignored
            -   (string):
                A corresponding item of metadata is located and used  to specify the box
            -   (tuple of 4 ints or floats):
                For each item in the tuple it is interpreted as follows:
                    -   (int):
                        A pixel coordinate in either the x or y direction
                    -   (float 0.0-1.0):
                        A fraction of the width or height in from the left, right, top, bottom sides
                    -   (float > 1.0):
                        Is rounded to the nearest integer and used a pixel coordinate.
                    -   None:
                        The extent of the image is used.

    Example:
        a=ImageFile(np.arange(12).reshape(3,4))

        a.crop(1,3,None,None)
    """
    box = self._box(*args, **kargs)
    ret = self[box]
    if "copy" in kargs.keys() and kargs["copy"]:
        ret = ret.clone
    return ret


def dtype_limits(self, clip_negative=True):
    """Return intensity limits, i.e. (min, max) tuple, of the image's dtype.

    Args:
        image(ndarray):
            Input image.
        clip_negative(bool):
            If True, clip the negative range (i.e. return 0 for min intensity)
            even if the image dtype allows negative values.

    Returns:
        (imin, imax : tuple): Lower and upper intensity limits.
    """
    if clip_negative is None:
        clip_negative = True
    imin, imax = dtype_range[self.dtype.type]
    if clip_negative:
        imin = 0
    return imin, imax


@keep_return_type
def asarray(self):
    """Provide a consistent way to get at the underlying array data in both ImageArray and ImageFile objects."""
    return self


def asfloat(self, normalise=True, clip=False, clip_negative=False):
    """Return the image converted to floating point type.

    If currently an int type and normalise then floats will be normalised
    to the maximum allowed value of the int type.
    If currently a float type then no change occurs.
    If clip then clip values outside the range -1,1
    If clip_negative then further clip values to range 0,1

    Keyword Arguments:
        normalise(bool):
            normalise the image to the max value of current int type
        clip_negative(bool):
            clip negative intensity to 0
    """
    if self.dtype.kind == "f":
        ret = self
    else:
        ret = self.convert(dtype=np.float64, normalise=normalise).view(type(self))  # preserve metadata
        tmp = metadataObject.__new__(metadataObject)
        for k, v in tmp.__dict__.items():
            if k not in ret.__dict__:
                ret.__dict__[k] = v
        c = self.clone  # copy formatting and apply to new array
        for k, v in c._optinfo.items():
            setattr(ret, k, v)
    if clip or clip_negative:
        ret = ret.clip_intensity(clip_negative=clip_negative)
    return ret


def clip_intensity(self, clip_negative=False, limits=None):
    """Clip intensity outside the range -1,1 or 0,1.

    Keyword Arguments:
        clip_negative(bool):
            if True clip to range 0,1 else range -1,1
        limits (low,high):
            Clip the intensity between low and high rather than zero and 1.

    Ensure data range is -1 to 1 or 0 to 1 if clip_negative is True.

    """
    if limits is None:
        dl = self.dtype_limits(clip_negative=clip_negative)
    else:
        dl = list(limits)
    np.clip(self, dl[0], dl[1], out=self)
    return self


def asint(self, dtype=np.uint16):
    """Convert the image to unsigned integer format.

    May raise warnings about loss of precision.
    """
    if self.dtype.kind == "f" and (np.max(self) > 1 or np.min(self) < -1):
        self = self.normalise()
    ret = self.convert(dtype)
    ret = ret.view(type(self))
    tmp = metadataObject.__new__(metadataObject)
    for k, v in tmp.__dict__.items():
        if k not in ret.__dict__:
            ret.__dict__[k] = v

    for k, v in self._optinfo.items():
        setattr(ret, k, v)
    return ret


def save(self, filename=None, **kargs):
    """Save the image into the file 'filename'.

    Args:
        filename (string, bool or None):
            Filename to save data as, if this is None then the current filename for the object is used
            If this is not set, then then a file dialog is used. If filename is False then a file dialog is forced.

    Keyword Args:
        fmt (string or list):
            format to save data as. 'tif', 'png' or 'npy' or a list of them. If not included will guess from
            filename.
        forcetype (bool):
            integer data will be converted to np.float32 type for saving. if forcetype then preserve and save as
            int type (will be unsigned).

    Notes:
        Metadata will be preserved in .png and .tif format.

        fmt can be 'png', 'npy', 'tif', 'tiff'  or a list of more than one of those.
        tif is recommended since metadata is lost in .npy format but data is
        converted to integer format for png so that definition cannot be
        saved.

    Since Stoner.Image is meant to be a general 2d array often with negative
    and floating point data this poses a problem for saving images. Images
    are naturally saved as 8 or more bit unsigned integer values representing colour.
    The only obvious way to save an image and preserve negative data
    is to save as a float32 tif. This has the advantage over the npy
    data type which cannot be opened by external programs and will not
    save metadata.
    """
    # Standard filename block
    if filename is None:
        filename = getattr(self, "filename", None)
    if filename is None or (isinstance(filename, bool) and not filename):
        # now go and ask for one
        filename = get_filedialog(what="file", filetypes=IMAGE_FILES)

    def_fmt = os.path.splitext(filename)[1][1:]  # Get a default format from the filename
    if def_fmt not in self.fmts:  # Default to png if nothing else
        def_fmt = "png"
    fmt = kargs.pop("fmt", [def_fmt])

    if not isinstance(fmt, list):
        fmt = [fmt]
    if set(fmt) & set(self.fmts) == set([]):
        raise ValueError(f"fmt must be {','.join(self.fmts)}")
    fmt = ["tiff" if f == "tif" else f for f in fmt]
    self.filename = filename
    for fm in fmt:
        saver = getattr(self, f"save_{fm}", "save_tif")
        if fm == "tiff":
            forcetype = kargs.pop("forcetype", False)
            saver(filename, forcetype)
        else:
            saver(filename)


def save_png(self, filename):
    """Save the ImageArray with metadata in a png file.

    This can only save as 8bit unsigned integer so there is likely
    to be a loss of precision on floating point data"""
    pngname = os.path.splitext(filename)[0] + ".png"
    meta = PngImagePlugin.PngInfo()
    info = self.metadata.export_all()
    info = [(i.split("=")[0], "=".join(i.split("=")[1:])) for i in info]
    for k, v in info:
        meta.add_text(k, v)
    s = (self - self.min()) * 256 / (self.max() - self.min())
    im = Image.fromarray(s.astype("uint8"), mode="L")
    im.save(pngname, pnginfo=meta)


def save_npy(self, filename):
    """Save the ImageArray as a numpy array."""
    npyname = os.path.splitext(filename)[0] + ".npy"
    np.save(npyname, np.array(self))


def save_tiff(self, filename, forcetype=False):
    """Save the ImageArray as a tiff image with metadata.

    Args:
        filename (str):
            Filename to save file as.

    Keyword Args:
        forcetype(bool):
            (deprecated) if forcetype then preserve data type as best as possible on save.
            Otherwise we let the underlying pillow library choose the best data type.

    Note:
        PIL can save in modes "L" (8bit unsigned int), "I" (32bit signed int),
        or "F" (32bit signed float). In general max info is preserved for "F"
        type so if forcetype is not specified then this is the default. For
        boolean type data mode "L" will suffice and this is chosen in all cases.
        The type name is added as a string to the metadata before saving.

    """
    from PIL.TiffImagePlugin import ImageFileDirectory_v2
    import json

    dtype = np.dtype(self.dtype).name  # string representation of dtype we can save
    self["ImageArray.dtype"] = dtype  # add the dtype to the metadata for saving.
    if forcetype:  # PIL supports uint8, int32 and float32, try to find the best match
        if self.dtype == np.uint8 or self.dtype.kind == "b":  # uint8 or boolean
            im = Image.fromarray(self, mode="L")
        elif self.dtype.kind in ["i", "u"]:
            im = Image.fromarray(self.astype("int32"), mode="I")
        else:  # default to float32
            im = Image.fromarray(self.astype(np.float32), mode="F")
    else:
        if (
            self.dtype.kind == "b"
        ):  # boolean we're not going to lose data by saving as unsigned int	            im = Image.fromarray(self)
            im = Image.fromarray(self, mode="L")
        else:  # try to convert everything else to float32 which can has maximum preservation of info
            try:
                im = Image.fromarray(self)
            except TypeError:
                im = Image.fromarray(self.astype("float32"))
    ifd = ImageFileDirectory_v2()
    ifd[270] = json.dumps(
        {"type": type(self).__name__, "module": type(self).__module__, "metadata": self.metadata.export_all()}
    )
    ext = os.path.splitext(filename)[1]
    if ext in [".tif", ".tiff"]:  # ensure extension is preserved in save
        pass
    else:  # default to tiff
        ext = ".tiff"
    tiffname = os.path.splitext(filename)[0] + ext
    im.save(tiffname, tiffinfo=ifd)



"""Stoner.HDF5 - Defines classes that use the hdf5 file format to store data on disc.

Classes include

* HDF5File - A :py:class:`Stoner.Code.DataFile` subclass that can save and load data from hdf5 files
* HDF5Folder - A :py:class:`Stoner.Folders.DataFolder` subclass that can save and load data from a single hdf5 file

It is only necessary to import this module for the subclasses of :py:class:`Stoner.Core.DataFile` to become available
to :py:class:`Stoner.Core.Data`.

"""

__all__ = ["HDF5File", "HDF5Folder", "HGXFile", "SLS_STXMFile", "STXMImage", "HDFFileManager"]
import importlib
import os.path as path
import os
from copy import deepcopy

import h5py
import numpy as np

from .compat import string_types, bytes2str, get_filedialog, path_types, str2bytes
from .Core import StonerLoadError, DataFile
from .folders import DataFolder
from .Image.core import ImageFile, ImageArray
from .core.utils import copy_into


def get_hdf_loader(f, default_loader=lambda *args, **kargs: None):
    """Look inside the open hdf file for details of what class to use to read this group.

    Args:
        f (h5py.File, h5py.Group):
            Open hdf file to look for a type attribute that gives the class to use to read this group.

    Returns:
        (callable):
            Callable function that can produce an object of an appropriate class.
    """
    if "type" not in f.attrs:
        raise StonerLoadError("HDF5 Group does not specify the type attribute used to check we can load it.")
    typ = bytes2str(f.attrs.get("type", ""))
    if (typ not in globals() or not isinstance(globals()[typ], type)) and "module" not in f.attrs:
        raise StonerLoadError(
            "HDF5 Group does not specify a recognized type and does not specify a module to use to load."
        )

    if "module" in f.attrs:
        mod = importlib.import_module(bytes2str(f.attrs["module"]))
        cls = getattr(mod, typ)
        return getattr(cls, "read_hdf5", default_loader)
    return getattr(globals()[typ], "read_hdf5", default_loader)


class HDFFileManager:
    """Context manager for HDF5 files."""

    def __init__(self, filename, mode="r"):
        """Initialise context handler.

        Works out the filename and group in cases the input flename includes a path to a sub group.

        Checks the file is actually an h4py file that is openable with the given mode.
        """
        self.mode = mode
        self.handle = None
        self.file = None
        self.close = True
        self.group = ""
        # Handle the case we're passed an already open h5py object
        if not isinstance(filename, path_types) or mode == "w":  # passed an already open h5py object
            self.filename = filename
            return
        # Here we deal with a string or path filename
        parts = str(filename).split(os.path.sep)
        bits = len(parts)
        for ix in range(bits):
            testname = "/".join(parts[: bits - ix])
            if path.exists(testname):
                filename = testname
                break

        try:
            if not mode.startswith("w"):
                with h5py.File(filename, "r"):
                    pass
        except (IOError, OSError) as err:
            raise StonerLoadError(f"{filename} not at HDF5 File") from err
        self.filename = filename

    def __enter__(self):
        """Open the hdf file with given mode and navigate to the group."""
        if isinstance(self.filename, (h5py.File, h5py.Group)):  # passed an already open h5py object
            self.handle = self.filename
            if isinstance(self.filename, h5py.Group):
                self.file = self.filename.file
            else:
                self.file = self.filename
            self.close = False
        elif isinstance(self.filename, path_types):  # Passed something to open
            handle = h5py.File(self.filename, self.mode)
            self.file = handle
            for grp in self.group.split("/"):
                if grp.strip() != "":
                    handle = handle[grp]
            self.handle = handle
        else:
            raise StonerLoadError("Note a resource that can be handled with HDF")
        return self.handle

    def __exit__(self, _type, _value, _traceback):
        """Ensure we close the hdf file no matter what."""
        if self.file is not None and self.close:
            self.file.close()


class HDF5File(DataFile):
    """A sub class of DataFile that sores itself in a HDF5File or group.

    Args:
        args (tuple):
            Supplied arguments, only recognises one though !
        kargs (dict):
            Dictionary of keyword arguments

    If the first non-keyword argument is not an h5py File or Group then
    initialises with a blank parent constructor and then loads data, otherwise,
    calls parent constructor.

    Datalayout is dead simple, the numerical data is in a dataset called *data*,
    metadata are attributes of a group called *metadata* with the keynames being the
    full name + typehint of the standard DataFile metadata dictionary
    *column_headers* are an attribute of the root file/group
    *filename* is set from either an attribute called filename, or from the
    group name or from the hdf5 filename.
    The root has an attribute *type* that must by 'HDF5File' otherwise the load routine
    will refuse to load it. This is to try to avoid loading rubbish from random hdf files.
    """

    priority = 16
    compression = "gzip"
    compression_opts = 6
    patterns = ["*.hdf", "*.hf5"]
    mime_type = ["application/x-hdf", "application/x-hdf5"]

    def __init__(self, *args, **kargs):
        """Initialise with an h5py.File or h5py.Group."""
        if args and isinstance(args[0], (h5py.File, h5py.Group)):
            args = list(args)
            grp = args.pop(0)
        else:
            grp = None
        super().__init__(*args, **kargs)
        if grp is not None:
            self._load(grp, **kargs)

    def _load(self, filename, *args, **kargs):
        """Load data from a hdf5 file.

        Args:
            h5file (string or h5py.Group):
                Either a string or an h5py Group object to load data from

        Returns:
            self:
                This object after having loaded the data
        """
        if filename is None or not filename:
            self.get_filename("r")
            filename = self.filename
        else:
            self.filename = filename
        with HDFFileManager(self.filename, "r") as f:
            loader = get_hdf_loader(f, default_loader=HDF5File.read_hdf)
            ret = loader(f, *args, instance=self, **kargs)
            if isinstance(ret, DataFile):
                copy_into(ret, self)
                return self
            return ret

    @classmethod
    def read_hdf(cls, filename, *args, **kargs):  # pylint: disable=unused-argument
        """Create a new HDF5File from an actual HDF file."""
        self = kargs.pop("instance", cls())
        if filename is None or not filename:
            self.get_filename("r")
            filename = self.filename
        with HDFFileManager(filename, "r") as f:
            data = f["data"]
            if np.prod(np.array(data.shape)) > 0:
                self.data = data[...]
            else:
                self.data = [[]]
            metadata = f.require_group("metadata")
            typehints = f.get("typehints", None)
            if not isinstance(typehints, h5py.Group):
                typehints = dict()
            else:
                typehints = typehints.attrs
            if "column_headers" in f.attrs:
                self.column_headers = [bytes2str(x) for x in f.attrs["column_headers"]]
                if isinstance(self.column_headers, string_types):
                    self.column_headers = self.metadata.string_to_type(self.column_headers)
                self.column_headers = [bytes2str(x) for x in self.column_headers]
            else:
                raise StonerLoadError("Couldn't work out where my column headers were !")
            for i in sorted(metadata.attrs):
                v = metadata.attrs[i]
                t = typehints.get(i, "Detect")
                if (
                    isinstance(v, string_types) and t != "Detect"
                ):  # We have typehints and this looks like it got exported
                    self.metadata[f"{i}{{{t}}}".strip()] = f"{v}".strip()
                else:
                    self[i] = metadata.attrs[i]
            if isinstance(f, h5py.Group):
                if f.name != "/":
                    self.filename = os.path.join(f.file.filename, f.name)
                else:
                    self.filename = os.path.realpath(f.file.filename)
            else:
                self.filename = os.path.realpath(f.filename)
        return self

    def save(self, filename=None, **kargs):
        """Write the current object into  an hdf5 file or group within a file.

        The data is written in afashion that is compatible with being loaded in again.

        Args:
            filename (string or h5py.Group):
                Either a string, of h5py.File or h5py.Group object into which to save the file. If this is a string,
                the corresponding file is opened for writing, written to and save again.

        Returns
            A copy of the object
        """
        return self.to_hdf(filename, **kargs)  # Just a pass through to our own to_hdf method

    def to_hdf(self, filename=None, **kargs):  # pylint: disable=unused-argument
        """Write the current object into  an hdf5 file or group within a file.

        Writes the data in afashion that is compatible with being loaded in again.

        Args:
            filename (string or h5py.Group):
                Either a string, of h5py.File or h5py.Group object into which to save the file. If this is a string,
                the corresponding file is opened for writing, written to and save again.

        Returns
            A copy of the object
        """
        if filename is None:
            filename = self.filename
        if filename is None or (isinstance(filename, bool) and not filename):  # now go and ask for one
            filename = self.__file_dialog("w")
            self.filename = filename
        if isinstance(filename, path_types):
            mode = "r+" if os.path.exists(filename) else "w"
        else:
            mode = ""
        with HDFFileManager(filename, mode) as f:
            f.require_dataset(
                "data",
                data=self.data,
                shape=self.data.shape,
                dtype=self.data.dtype,
                compression=self.compression,
                compression_opts=self.compression_opts,
            )
            metadata = f.require_group("metadata")
            typehints = f.require_group("typehints")
            for k in self.metadata:
                try:
                    typehints.attrs[k] = self.metadata._typehints[k]
                    metadata.attrs[k] = self[k]
                except TypeError:
                    # We get this for trying to store a bad data type - fallback to metadata export to string
                    parts = self.metadata.export(k).split("=")
                    metadata.attrs[k] = "=".join(parts[1:])
            f.attrs["column_headers"] = [str2bytes(x) for x in self.column_headers]
            f.attrs["filename"] = self.filename
            f.attrs["type"] = type(self).__name__
            f.attrs["module"] = type(self).__module__
        return self


class HGXFile(DataFile):
    """A subclass of DataFile for reading GenX HDF Files.

    These files typically have an extension .hgx. This class has been based on a limited sample
    of hgx files and so may not be sufficiently general to handle all cases.
    """

    priority = 16
    pattern = ["*.hgx"]
    mime_type = ["application/x-hdf", "application/x-hdf5"]

    def _load(self, filename, *args, **kargs):
        """Load a GenX HDF file.

        Args:
            filename (string or bool):
                File to load. If None then the existing filename is used, if False, then a file dialog will be used.

        Returns:
            A copy of the itself after loading the data.
        """
        self.seen = []
        if filename is None or not filename:
            self.get_filename("r")
        else:
            self.filename = filename

        with HDFFileManager(self.filename, "r") as f:
            try:
                if "current" in f and "config" in f["current"]:
                    pass
                else:
                    raise StonerLoadError("Looks like an unexpected HDF layout!.")
                self.scan_group(f["current"], "")
                self.main_data(f["current"]["data"])
            except IOError as err:
                raise StonerLoadError("Looks like an unexpected HDF layout!.") from err
        return self

    def scan_group(self, grp, pth):
        """Recursively list HDF5 Groups."""
        if pth in self.seen:
            return None
        self.seen.append(pth)
        if not isinstance(grp, h5py.Group):
            return None
        if self.debug:
            if self.debug:
                print(f"Scanning in {pth}")
        for x in grp:
            if pth == "":
                new_pth = x
            else:
                new_pth = pth + "." + x
            if pth == "" and x == "data":  # Special case for main data
                continue
            if isinstance(grp[x], type(grp)):
                self.scan_group(grp[x], new_pth)
            elif isinstance(grp[x], h5py.Dataset):
                y = grp[x][...]
                self[new_pth] = y
        return None

    def main_data(self, data_grp):
        """Work through the main data group and build something that looks like a numpy 2D array."""
        if not isinstance(data_grp, h5py.Group) or data_grp.name != "/current/data":
            raise StonerLoadError("HDF5 file not in expected format")
        root = data_grp["datasets"]
        for ix in root:  # Hack - iterate over all items in root, but actually data is in Groups not DataSets
            dataset = root[ix]
            if isinstance(dataset, h5py.Dataset):
                continue
            x = dataset["x"][...]
            y = dataset["y"][...]
            e = dataset["error"][...]
            self &= x
            self &= y
            self &= e
            self.column_headers[-3] = bytes2str(dataset["x_command"][()])
            self.column_headers[-2] = bytes2str(dataset["y_command"][()])
            self.column_headers[-1] = bytes2str(dataset["error_command"][()])
            self.column_headers = [str(ix) for ix in self.column_headers]


class HDF5FolderMixin:
    """Provides a method to load and save data from a single HDF5 file with groups.

    See :py:class:`Stoner.Folders.DataFolder` for documentation on constructor.

    Datalayout consistns of sub-groups that are either instances of HDF5Files (i.e. have a type attribute that
    contains 'HDF5File') or are themsleves HDF5Folder instances (with a type attribute that reads 'HDF5Folder').
    """

    def __init__(self, *args, **kargs):
        """Initialise the File aatribute."""
        self.File = None
        super().__init__(*args, **kargs)

    def __getter__(self, name, instantiate=True):
        """Load the specified name from a file on disk.

        Parameters:
            name (key type):
                The canonical mapping key to get the dataObject. By default
                the baseFolder class uses a :py:class:`regexpDict` to store objects in.

        Keyword Arguments:
            instantiate (bool):
                If True (default) then always return a :py:class:`Stoner.Core.Data` object. If False,
                the __getter__ method may return a key that can be used by it later to actually get the
                :py:class:`Stoner.Core.Data` object.

        Returns:
            (metadataObject):
                The metadataObject
        """
        try:
            return super().__getter__(name, instantiate=instantiate)
        except (AttributeError, IndexError, KeyError, OSError, IOError) as err:
            if self.debug:
                print(err)
        names = list(os.path.split(name))
        if names[0] == "/":  # Prune leading ./
            names = names[1:]
        if self.File is None:
            closeme = True
            self.File = h5py.File(self.directory, "r+")
        else:
            closeme = False
        grp = self.File
        while len(names) > 0:
            next_group = names.pop(0)
            if next_group not in grp:
                raise IOError(f"Cannot find {name} in {self.File.filename}")
            grp = grp[next_group]
        tmp = self.loader(grp)
        tmp.filename = grp.name
        tmp = self.on_load_process(tmp)
        tmp = self._update_from_object_attrs(tmp)
        self.__setter__(name, tmp)
        if closeme:
            self.File.close()
            self.File = None
        return tmp

    def _dialog(self, message="Select Folder", new_directory=True, mode="r+"):
        """Create a file dialog box for working with.

        Args:
            message (string):
                Message to display in dialog
            new_file (bool):
                True if allowed to create new directory

        Returns:
            A directory to be used for the file operation.
        """
        file_wildcard = "hdf file (*.hdf5)|*.hdf5|Data file (*.dat)|\
        *.dat|All files|*"

        if mode == "r":
            mode2 = "Open HDF file"
        elif mode == "w":
            mode2 = "Save HDF file as"

        dlg = get_filedialog(
            "file", title=mode2, filetypes=file_wildcard, message=message, mustexist=not new_directory
        )
        if len(dlg) != 0:
            self.directory = dlg
            self.File = h5py.File(self.directory, mode)
            self.File.close()
            return self.directory
        return None

    def _visit_func(self, name, obj):  # pylint: disable=unused-argument
        """Walker of the HDF5 tree."""
        if isinstance(obj, h5py.Group) and "type" in obj.attrs:
            cls = globals()[obj.attrs["type"]]
            if issubclass(self.loader, cls):
                self.__setter__(obj.name, obj.name)
            elif obj.attrs["type"] == "HDF5Folder" and self.recursive:
                self.groups[obj.name.split(path.sep)[-1]] = type(self)(
                    obj, pattern=self.pattern, type=self.type, recursive=self.recursive, loader=self.loader
                )

    def close(self):
        """Close the cirrent hd5 file."""
        if isinstance(self.File, h5py.File):
            self.File.close()
            self.File = None
        else:
            raise IOError("HDF5 File not open!")

    def getlist(self, recursive=None, directory=None, flatten=False):
        """Read the HDF5 File to construct a list of file HDF5File objects."""
        if recursive is None:
            recursive = self.recursive
        self.files = []
        self.groups = {}
        closeme = True
        for d in [directory, self.directory, self.File, True]:
            if isinstance(d, bool) and d:
                d = self._dialog()
            directory = d
            if d is not None:
                break
        if directory is None:
            return None
        if isinstance(directory, path_types):
            try:
                self.directory = directory
                directory = h5py.File(directory, "r+")
                self.File = directory
                closeme = True
            except OSError:
                return super().getlist(recursive, directory, flatten)
        elif isinstance(directory, h5py.File) or isinstance(directory, h5py.Group):  # Bug out here
            self.File = directory.file
            self.directory = self.File.filename
            closeme = False
        # At this point directory contains an open h5py.File object, or possibly a group
        directory.visititems(self._visit_func)
        if flatten:
            self.flatten()
        if closeme:
            self.File.close()
            self.File = None
        return self

    def save(self, root=None):
        """Save a load of files to a single HDF5 file, creating groups as it goes.

        Keyword Arguments:
            root (string):
                The name of the HDF5 file to save to if set to None, will prompt for a filename.

        Return:
            A list of group paths in the HDF5 file
        """
        closeme = False
        if root is None and isinstance(self.File, h5py.File):
            root = self.File
        elif root is None and not isinstance(self.File, h5py.File):
            root = h5py.File(self.directory, mode="w")
            self.File = root
            closeme = True
        if root is None or (isinstance(root, bool) and not root):
            # now go and ask for one
            root = self._dialog()
            root = h5py.File(root, mode="a")
            self.File = root
            closeme = True
        if isinstance(root, path_types):
            mode = "r+" if path.exists(root) else "w"
            root = h5py.File(root, mode)
            self.File = root
            closeme = True
        if not isinstance(root, (h5py.File, h5py.Group)):
            raise IOError("Can't save Folder without an HDF5 file or Group!")
        # root should be an open h5py file
        root.attrs["type"] = "HDF5Folder"
        for ix, obj in enumerate(self):
            name = os.path.basename(getattr(obj, "filename", f"obj-{ix}"))
            if name not in root:
                root.create_group(name)
            name = root[name]
            self.loader(obj).save(name)
        for grp in self.groups:
            if grp not in root:
                root.create_group(grp)
            self.groups[grp].save(root[grp])

        if closeme and self.File is not None:
            self.File.close()
            self.File = None
        return self


class HDF5Folder(HDF5FolderMixin, DataFolder):
    """Just enforces the loader attriobute to be an HDF5File."""

    def __init__(self, *args, **kargs):
        """Ensure the loader routine is set for HDF5Files."""
        self.loader = HDF5File
        super().__init__(*args, **kargs)


class SLS_STXMFile(DataFile):
    """Load images from the Swiss Light Source Pollux beamline."""

    priority = 16
    compression = "gzip"
    compression_opts = 6
    patterns = ["*.hdf"]
    mime_type = ["application/x-hdf", "application/x-hdf5"]

    def _load(self, filename, *args, **kargs):
        """Load data from the hdf5 file produced by Pollux.

        Args:
            h5file (string or h5py.Group):
                Either a string or an h5py Group object to load data from

        Returns:
            itself after having loaded the data
        """
        if filename is None or not filename:
            self.get_filename("r")
            filename = self.filename
        else:
            self.filename = filename
        if isinstance(filename, path_types):  # We got a string, so we'll treat it like a file...
            try:
                f = h5py.File(filename, "r")
            except IOError as err:
                raise StonerLoadError(f"Failed to open {filename} as a n hdf5 file") from err
        elif isinstance(filename, h5py.File) or isinstance(filename, h5py.Group):
            f = filename
        else:
            raise StonerLoadError(f"Couldn't interpret {filename} as a valid HDF5 file or group or filename")
        items = [x for x in f.items()]
        if len(items) == 1 and items[0][0] == "entry1":
            group1 = [x for x in f["entry1"]]
            if "definition" in group1 and bytes2str(f["entry1"]["definition"][0]) == "NXstxm":  # Good HDF5
                pass
            else:
                raise StonerLoadError("HDF5 file lacks single top level group called entry1")
        else:
            raise StonerLoadError("HDF5 file lacks single top level group called entry1")
        root = f["entry1"]
        data = root["counter0"]["data"]
        if np.prod(np.array(data.shape)) > 0:
            self.data = data[...]
        else:
            self.data = [[]]
        self.scan_meta(root)
        if "file_name" in f.attrs:
            self["original filename"] = f.attrs["file_name"]
        elif isinstance(f, h5py.Group):
            self["original filename"] = f.name
        else:
            self["original filename"] = f.file.filename

        if isinstance(filename, path_types):
            f.file.close()
        self["Loaded from"] = self.filename

        if "instrument.sample_x.data" in self.metadata:
            self.metadata["actual_x"] = self.metadata["instrument.sample_x.data"].reshape(self.shape)
        if "instrument.sample_y.data" in self.metadata:
            self.metadata["actual_y"] = self.metadata["instrument.sample_y.data"].reshape(self.shape)
        self.metadata["sample_x"], self.metadata["sample_y"] = np.meshgrid(
            self.metadata["counter0.sample_x"], self.metadata["counter0.sample_y"]
        )
        if "control.data" in self.metadata:
            self.metadata["beam current"] = ImageArray(self.metadata["control.data"].reshape(self.data.shape))
            self.metadata["beam current"].metadata = self.metadata
        self.data = self.data[::-1]
        return self

    def scan_meta(self, group):
        """Scan the HDF5 Group for attributes and datasets and sub groups and recursively add them to the metadata."""
        root = ".".join(group.name.split("/")[2:])
        for name, thing in group.items():
            parts = thing.name.split("/")
            name = ".".join(parts[2:])
            if isinstance(thing, h5py.Group):
                self.scan_meta(thing)
            elif isinstance(thing, h5py.Dataset):
                if thing.ndim > 1:
                    continue
                if np.prod(thing.shape) == 1:
                    self.metadata[name] = thing[0]
                else:
                    self.metadata[name] = thing[...]
        for attr in group.attrs:
            self.metadata[f"{root}.{attr}"] = group.attrs[attr]


class STXMImage(ImageFile):
    """An instance of KerrArray that will load itself from a Swiss Light Source STXM image."""

    # pylint: disable=no-member

    _reduce_metadata = False

    _patterns = ["*.hdf5", "*.hdf"]
    mime_type = ["application/x-hdf", "application/x-hdf5"]
    priority = 16

    def __init__(self, *args, **kargs):
        """Initialise and load a STXM image produced by Pollux.

        Keyword Args:
            regrid (bool):
                If set True, the gridimage() method is automatically called to re-grid the image to known coordinates.
        """
        regrid = kargs.pop("regrid", False)
        bcn = kargs.pop("bcn", False)
        if len(args) > 0 and isinstance(args[0], path_types):
            d = SLS_STXMFile(args[0])
            args = args[1:]
            super().__init__(*args, **kargs)
            self.image = d.data
            self.metadata = deepcopy(d.metadata)
            self.filename = d.filename
        elif len(args) > 0 and isinstance(args[0], ImageFile):
            src = args[0]
            args = args[1:]
            super().__init__(src.image, *args, **kargs)
        elif len(args) > 0 and isinstance(args[0], ImageArray):
            src = args[0]
            args = args[1:]
            super().__init__(src, *args, **kargs)
        else:
            super().__init__(*args, **kargs)
        if isinstance(regrid, tuple):
            self.gridimage(*regrid)
        elif isinstance(regrid, dict):
            self.gridimage(**regrid)
        elif regrid:
            self.gridimage()
        if bcn:
            if regrid:
                self.metadata["beam current"] = self.metadata["beam current"].gridimage()
            self.image /= self["beam current"]

    @property
    def polarization(self):
        """Return the sign of the polarization used for the image.

        Returns:
            +1,-1 or 0 for positive, negative or unpolariused light used for the image.
        """
        return np.sign(self.get("collection.polarization.value", 0))

    def _load(self, filename, *args, **kargs):
        """Pass through to SLS_STXMFile._load."""
        self.__init__(filename, *args, **kargs)
        return self



"""Example of using scale to overlap data."""

# pylint: disable=invalid-name, no-member
from numpy import linspace, sin, exp, pi, column_stack
from numpy.random import normal, seed
import matplotlib as mpl
from tabulate import tabulate

from Stoner import Data

seed(3)  # Just fix the random numbers to stop optimizer warnings
mpl.rc("text", usetex=True)


x = linspace(0, 10 * pi, 201)
x2 = x * 1.5 + 0.23
y = 10 * exp(-x / (2 * pi)) * sin(x) + normal(size=len(x), scale=0.1)
y2 = 3 * exp(-x / (2 * pi)) * sin(x) - 1 + normal(size=len(x), scale=0.1)

d = Data(x, y, column_headers=["Time", "Signal 1"], setas="xy")
d2 = Data(x2, y2, column_headers=["Time", "Signal 2"], setas="xy")

d.plot(label="1$^\\mathrm{st}$ signal")
d2.plot(figure=d.fig, label="2$^\\mathrm{nd}$ signal")
d3 = d2.scale(d, header="Signal 2 scaled", xmode="affine")
d3.plot(figure=d.fig, label="1$^\\mathrm{st}$ scaled signals")
d3["test"] = linspace(1, 10, 10)
txt = tabulate(d3["Transform"], floatfmt=".2f", tablefmt="grid")
d3.text(10, 4, "Transform\n{}".format(txt), fontdict={"size": "x-small"})

np_data = column_stack((x2, y2))
d4 = d.scale(
    np_data, header="Signal 2 scaled", xmode="affine", use_estimate=True
)
d4.plot(figure=d.fig, label="2$^\\mathrm{nd}$ scaled signal")
d4.ylim = (-7, 9)
txt = tabulate(d4["Transform"], floatfmt=".2f", tablefmt="grid")
d4.text(10, -7, "Transform\n{}".format(txt), fontdict={"size": "x-small"})

d4.title = "Scaling Example"



"""Test Weak-localisation fitting."""

# pylint: disable=invalid-name
from copy import copy

from numpy import linspace, ones_like
from numpy.random import normal

from Stoner import Data
from Stoner.analysis.fitting.models.magnetism import (
    KittelEquation,
    kittelEquation,
)
from Stoner.plot.formats import TexEngFormatter

B = linspace(1e3, 5e4, 51)
params = [2.2, 1e5, 2e2]
G = kittelEquation(B, *params) + normal(size=len(B), scale=5e7)
dG = ones_like(B) * 5e7

d = Data(
    B,
    G,
    dG,
    setas="xye",
    column_headers=["Field $Oe$", r"$\nu (Hz)$", r"\delta $\nu (Hz)$"],
)

d.curve_fit(kittelEquation, p0=copy(params), result=True, header="curve_fit")

fit = KittelEquation()
p0 = fit.guess(G, x=B)

d.lmfit(fit, p0=p0, result=True, header="lmfit")

d.setas = "xyeyy"
d.plot(fmt=["r.", "b-", "g-"])

d.annotate_fit(
    kittelEquation,
    x=0.5,
    y=0.25,
    fontdict={"size": "x-small", "color": "blue"},
    mode="eng",
)
d.annotate_fit(
    KittelEquation,
    x=0.5,
    y=0.05,
    fontdict={"size": "x-small", "color": "green"},
    prefix="KittelEquation",
    mode="eng",
)
d.title = "Kittel Fit"
d.fig.gca().xaxis.set_major_formatter(TexEngFormatter())
d.fig.gca().yaxis.set_major_formatter(TexEngFormatter())



"""Example of nDimArrhenius Fit."""

# pylint: disable=invalid-name
from numpy import linspace
from numpy.random import normal

from Stoner import Data
from Stoner.analysis.fitting.models.thermal import nDimArrhenius, NDimArrhenius

# Make some data
T = linspace(50, 500, 101)
R = nDimArrhenius(T + normal(size=len(T), scale=5.0, loc=1.0), 1e6, 0.5, 2)
d = Data(T, R, setas="xy", column_headers=["T", "Rate"])

# Curve_fit on its own
d.curve_fit(nDimArrhenius, p0=[1e6, 0.5, 2], result=True, header="curve_fit")
d.setas = "xyy"
d.plot(fmt=["r.", "b-"])
d.annotate_fit(nDimArrhenius, x=0.25, y=0.3, mode="eng")

# lmfit using lmfit guesses
fit = NDimArrhenius()
p0 = fit.guess(R, x=T)
d.lmfit(fit, result=True, header="lmfit")
d.setas = "x..y"
d.plot(fmt="g-")
d.annotate_fit(
    NDimArrhenius, x=0.25, y=0.05, prefix="NDimArrhenius", mode="eng"
)

d.title = "n-D Arrhenius Test Fit"
d.ylabel = "Rate"
d.xlabel = "Temperature (K)"



# -*- coding: utf-8 -*-
"""Tests for Stoner.tools.classes"""


import pytest
from Stoner.tools import classes
from Stoner import Options


def test_typedList_create():
    tl = classes.typedList()
    assert tl._type == str, "Empty typedList should default to str"
    tl = classes.typedList(int)
    assert tl._type == int and len(tl) == 0, "typedList with single type constructor failed"
    tl = classes.typedList(int, [1, 2, 3, 4, 5])
    assert tl._type == int and len(tl) == 5, "typedList with type and iterator constructor failed"
    try:
        tl = classes.typedList(int, 1, 2, 3)
    except SyntaxError:
        pass
    else:
        assert False, "typedList with too many arguments didn't throw error"
    try:
        tl = classes.typedList(int, ["A", 1, 2, 3])
    except TypeError:
        pass
    else:
        assert False, "typedList with with bad types didn't throw error"


def test_typedList_operators():
    tl1 = classes.typedList(int, (1, 2, 3))
    tl2 = classes.typedList(int, (4, 5, 6))
    tl3 = classes.typedList(float, (1.1, 1.2, 1.3))
    assert len(tl1 + tl2) == 6, "Failed to add similar typed lists"
    try:
        _ = tl1 + tl3
    except TypeError:
        pass
    else:
        assert False, "Adding unlike typed lists didn't raise TyperError"
    tl1 += tl2
    assert len(tl1) == 6, "In place add of typedList failed"
    try:
        tl1 += 2
    except TypeError:
        pass
    else:
        assert False, "Adding non list to a typed lists didn't raise TyperError"
    assert [True, 2] + tl3 == [True, 2, 1.1, 1.2, 1.3], "Right adding typedList failed."
    tl1 = classes.typedList(int, (1, 2, 3))
    tl2 = classes.typedList(int, (1, 2, 3))
    assert tl1 == tl2, "typedList equality check failed"

    del tl1[0]
    assert tl1 == [2, 3], "Failed to delete and item from the list"
    assert tl1[0] == 2, "typedList get item failed"
    assert repr(tl1) == "[2, 3]", "typedList representation failed"
    try:
        tl1[[0, 1]] = 3
    except TypeError:
        pass
    else:
        assert (
            False
        ), "typedList setitem with mismatch between iterable key and scalar value failed to rasise a TypeError"
    try:
        tl1[[0, 1]] = 3.0
    except TypeError:
        pass
    else:
        assert False, "typedList setitem with from type of value failed to rasise a TypeError"


def test_typedList_methods():
    tl1 = classes.typedList(int, (1, 2, 3))
    try:
        tl1.extend(1)
    except TypeError:
        pass
    else:
        assert False, "typedList failed to raise typeError when extending with a scale"
    assert tl1.index(2) == 1, "typedList index failed to give right answer"
    tl1.insert(0, 0)
    assert tl1 == [0, 1, 2, 3], "typedList insert at start failed."
    try:
        tl1.insert(0, "G")
    except TypeError:
        pass
    else:
        assert False, "typedList failed to raise typeError when inserting with the wrong type"


def test_Options():
    try:
        classes.get_option("Random")
    except IndexError:
        pass
    else:
        assert False, "get_option didn't raise an IndexError for a bad option name"
    try:
        classes.set_option("Random", False)
    except IndexError:
        pass
    else:
        assert False, "set_option didn't raise an IndexError for a bad option name"
    try:
        classes.set_option("no_figs", "Hi")
    except ValueError:
        pass
    else:
        assert False, "set_option didn't raise a ValueError for a bad option value"

    try:
        Options.Random = False
    except AttributeError:
        pass
    else:
        assert False, "Set Option attribute didn't raise an IndexError for a bad option name"
    try:
        Options.no_figs = "Hi"
    except ValueError:
        pass
    else:
        assert False, "Set Options attribute didn't raise a ValueError for a bad option value"
    Options.no_figs = False
    assert not Options.no_figs, "Setting Options attribute didn't stick"
    del Options.no_figs
    assert Options.no_figs, "Deleting Options attrkibute didn't clear option"
    assert dir(Options) == [
        "multiprocessing",
        "no_figs",
        "short_data_repr",
        "short_folder_rrepr",
        "short_img_repr",
        "short_repr",
        "threading",
        "warnings",
    ], "Directory of Options failed"
    opt_repr = (
        "Stoner Package Options\n~~~~~~~~~~~~~~~~~~~~~~\n"
        + "multiprocessing : False\nno_figs : True\nshort_data_repr : False\nshort_folder_rrepr : True\n"
        + "short_img_repr : True\nshort_repr : False\nthreading : False\nwarnings : False\n"
    )
    assert repr(Options) == opt_repr, "Representation of Options failed"


if __name__ == "__main__":
    pytest.main(["--pdb", __file__])



"""3D surface plot example."""

# pylint: disable=invalid-name
import numpy as np
import matplotlib.cm

from Stoner import Data

x, y, z = np.meshgrid(
    np.linspace(-2, 2, 21), np.linspace(-2, 2, 21), np.linspace(-2, 2, 21)
)
x = x.ravel()
y = y.ravel()
z = z.ravel()
u = np.sin(x * y * z)

p = Data(x, y, z, u, setas="xyzu", column_headers=["X", "Y", "Z"])

p.plot_voxels(cmap=matplotlib.cm.jet, visible=lambda x, y, z: x - y + z < 2.0)
p.set_box_aspect((1, 1, 1.0))  # Passing through to the current axes
p.title = "Voxel plot"



"""Plot data using multiple sub-plots."""

# pylint: disable=invalid-name
from Stoner import Data

p = Data("sample.txt", setas="xyy")
# Quick plot
p.plot(multiple="panels")
# Helps to fix layout !
p.set_layout_engine("tight")
p.draw()



# -*- coding: utf-8 -*-
"""Line and Box selection Tools for Images."""

import time
import numpy as np

import matplotlib.pyplot as plt
from scipy.optimize import minimize
from matplotlib.widgets import Cursor, RectangleSelector
from matplotlib.colors import to_rgba
from matplotlib.backend_bases import Event

from skimage import draw


def send_event(image, names, **kargs):
    """Make a fake event."""
    time.sleep(0.05)
    select = image._image._select
    event = Event("fake", select.fig.canvas)
    if not isinstance(names, list):
        names = [names]
    for name in names:
        for k, v in kargs.items():
            setattr(event, k, v)
        getattr(select, name)(event)


def _straight_ellipse(p, data):
    """A non-rotated ellipse."""
    xc, yc, a, b = p
    x, y = data.T
    t1 = (x - xc) ** 2 / a**2
    t2 = (y - yc) ** 2 / b**2
    return np.abs(np.sum(t1 + t2 - 1.0))


def _rotated_ellipse(p, data):
    """A non-rotated ellipse."""
    xc, yc, a, b, phi = p
    x, y = data.T
    t1 = ((x - xc) * np.cos(phi) + (y - yc) * np.sin(phi)) ** 2 / a**2
    t2 = ((x - xc) * np.sin(phi) - (y - yc) * np.cos(phi)) ** 2 / b**2
    return np.abs(np.sum(t1 + t2 - 1.0))


class LineSelect:
    """Show an Image and slow the user to draw a line on it using cursors."""

    def __init__(self):
        """Create the LineSelect object, display the image and register the hooks.

        The constructor will wait until the finished coordinates are set and then return
        [(x_start,y_start),(x_finish, y_finish)].
        """
        self.started = False
        self.finished = False
        self.ax = None
        self.fig = None
        self.crs = None
        self.mode = "xy"
        self.kargs = {}

    def __call__(self, image, **kargs):
        """Do the actual line selection.

        Args:
            image (ImageArray, ImageFile):
                The image to shopw to the user for the selection
            **kargs (mixed):
                Other keywords to pass to the line drawing.

        Returns:
            list:
                [(x_start,y_start),(x_finish,y_finish)] coordinates for the endpoints of the line

        """
        self.fig = image.imshow()
        self.ax = plt.gca()
        self.kargs = kargs
        image._select = self  # allows us to hook to the selector.
        self.crs = Cursor(self.ax)
        self.crs.connect_event("button_press_event", self.on_click)
        self.crs.connect_event("motion_notify_event", self.draw_line)
        self.fig.canvas.mpl_connect("key_press_event", self.keypress)
        while not self.finished:
            plt.pause(0.01)
        plt.close(self.fig.number)
        delattr(image, "_select")  # cleanup
        return [self.started, self.finished]

    def keypress(self, event):
        """Habndle key press events.

        Args:
            event (matplotlib event):
                The matplotlib event object

        Returns:
            None.

        Tracks the mode and adjusts the cursor display.
        """
        if event.key.lower() == "x":
            if self.mode == "x":
                self.mode = "xy"
            else:
                self.mode = "x"
        elif event.key.lower() == "y":
            if self.mode == "y":
                self.mode = "xy"
            else:
                self.mode = "y"
        self.crs.horizOn = "y" in self.mode
        self.crs.vertOn = "x" in self.mode

    def on_click(self, event):
        """Habndle mouse click events.

        Args:
            event (matplotlib event):
                The matplotlib event object

        Returns:
            None.

        Records with the starting or finishing coordinates of the line.
        """
        if self.mode == "x":
            y1, y2 = self.ax.get_ylim()
            x = event.xdata
            self.started = (x, y1)
            self.finished = (x, y2)
            return
        if self.mode == "y":
            x1, x2 = self.ax.get_xlim()
            y = event.ydata
            self.started = (x1, y)
            self.finished = (x2, y)
            return
        if not self.started:
            self.started = event.xdata, event.ydata
        else:
            self.finished = event.xdata, event.ydata

    def draw_line(self, event):
        """Handle the drawing of the selection line.

        Args:
            event (matplotlib event):
                The matplotlib event object

        Returns:
            None.

        Optiuonal line properties can be overridden by passing keyword parameters to the constructor.
        """
        if not self.started:  # Do nothing until we start
            return
        if len(self.ax.lines) > 2:  # Rremove the old line
            self.ax.lines[2].remove()

        self.kargs.setdefault("linewidth", 2)
        self.kargs.setdefault("linestyle", "dash")
        self.kargs.setdefault("")

        xc, yc = event.xdata, event.ydata
        xs, ys = self.started
        plt.plot([xs, xc], [ys, yc], "r-", linewidth=2)


class RegionSelect:
    """Show an Image and slow the user to select a rectangular section."""

    def __init__(self):
        """Create the LineSelect object, display the image and register the hooks.

        The constructor will wait until the finished coordinates are set and then return
        [(x_start,y_start),(x_finish, y_finish)].
        """
        self.p1 = False
        self.p2 = False
        self.finished = False
        self.ax = None
        self.fig = None
        self.select = None
        self.kargs = {}

    def __call__(self, image, **kargs):
        """Actuall do the region selection.

        Args:
            image (ImageArray, ImageFile):
                The image to shopw to the user for the selection
            **kargs (mixed):
                Other keywords to pass to the line drawing.

        Returns:
            list:
                The x and y range coordinates, in te order left-x, right-x, top-y, bottom-y

        """
        self.fig = image.imshow()
        plt.title("Click and drag to select region and press return")
        self.ax = plt.gca()
        self.kargs = kargs
        image._select = self  # allows us to hook to the selector.
        self.select = RectangleSelector(
            self.ax, self.on_select, button=[1], minspanx=5, minspany=5, useblit=True, interactive=True
        )
        self.fig.canvas.mpl_connect("key_press_event", self.finish)
        while not self.finished:
            plt.pause(0.01)
        plt.close(self.fig.number)
        delattr(image, "_select")  # cleanup
        return [
            min(self.p1[0], self.p2[0]),
            max(self.p1[0], self.p2[0]),
            min(self.p1[1], self.p2[1]),
            max(self.p1[1], self.p2[1]),
        ]

    def on_select(self, start, stop):
        """Habndle mouse click events.

        Args:
            start (matplotlib event):
                The matplotlib event object for the start of the selection
            stop (matplotlib event):
                The matplotlib event object for the start of the selection

        Returns:
            None.

        Records with the locations of the corners of the rectangular section.
        """
        self.p1 = (int(np.round(start.xdata)), int(np.round(start.ydata)))
        self.p2 = (int(np.round(stop.xdata)), int(np.round(stop.ydata)))
        self.select.active = True

    def finish(self, key_event):
        """Handle ley press events.

        Args:
            key_event (matplotlib event):
                The matplotlib event object

        Returns:
            None.

        Records with the Finish of the selection on <enter>.
        """
        if self.p1 and self.p2 and key_event.key == "enter":
            self.finished = True


class ShapeSelect:
    """Show an Image and slow the user to draw a line on it using cursors."""

    def __init__(self):
        """Create the LineSelect object, display the image and register the hooks.

        The constructor will wait until the finished coordinates are set and then return
        [(x_start,y_start),(x_finish, y_finish)].
        """
        self.invert = False
        self.ov_layer = None
        self.shape = None
        self.colour = "red"
        self.alpha = 0.5
        self.obj = "poly"
        self.finished = False
        self.vertices = []
        self.ax = None
        self.fig = None
        self.crs = None
        self.mode = "xy"
        self.kargs = {}

    def __call__(self, image, **kargs):
        """Do the work of the polygon selection.

        Args:
            image (ImageArray, ImageFile):
                The image to shopw to the user for the selection
            **kargs (mixed):
                Other keywords to pass to the drawing.

        Returns:
            mask_array:
                A boolean array of te same shape as the image that can be used as a mask.

        """
        self.fig = image.imshow()
        self.shape = image.shape
        self.ax = plt.gca()
        # Sort colours out
        self.colour = kargs.pop("colour", getattr(image.mask, "colour", "red"))
        self.alpha = kargs.pop("alpha", 0.5)
        self.colour = np.array(to_rgba(self.colour))
        self.colour[3] = self.alpha
        # Create overlay plot
        overlay = np.zeros(self.shape + (4,))
        self.ax.imshow(overlay)
        self.ov_layer = self.ax.images[-1]

        self.kargs = kargs
        image._select = self  # allows us to hook to the selector.
        self.crs = Cursor(self.ax)
        plt.title(self.draw_poly.instructions)
        plt.xlabel(
            "LMB: Select, RMB: remove, LMB-Dbl: finish, Esc: cancel\ni: invert,"
            + "c: circle/ellipse, r:rectangle, p:polygon"
        )
        plt.pause(0.1)
        bp_ev = self.crs.connect_event("button_press_event", self.on_click)
        kp_ev = self.crs.connect_event("motion_notify_event", self.draw)
        mm_ev = self.fig.canvas.mpl_connect("key_press_event", self.keypress)
        while not self.finished:
            plt.pause(0.01)
        delattr(image, "_select")  # cleanup
        plt.disconnect(bp_ev)
        plt.disconnect(kp_ev)
        plt.disconnect(mm_ev)
        plt.close(self.fig.number)
        return self.get_mask()

    def keypress(self, event):
        """Habndle key press events.

        Args:
            event (matplotlib event):
                The matplotlib event object

        Returns:
            None.

        Tracks the mode and adjusts the cursor display.
        """
        shapes = {"p": "poly", "c": "circle", "r": "rectangle"}
        if event.key.lower() in shapes:
            self.obj = shapes[event.key.lower()]
            plt.title(getattr(getattr(self, f"draw_{self.obj}", ""), "instructions", ""))
            return self.draw(event)
        if event.key.lower() == "i":
            self.invert = ~self.invert
            return self.draw(event)
        if event.key.lower() == "enter":
            self.vertices.append((event.xdata, event.ydata))
            event.button = 1
            if len(self.vertices) >= getattr(getattr(self, f"draw_{self.obj}", ""), "min_vertices", 0):
                event.dblclick = True
                return self.on_click(event)
            return self.draw(event)
        if event.key == "escape":
            self.vertices = []
            self.invert = False
            self.ov_layer.set_array(np.ones(self.shape + (4,)))
            self.finished = True
            return self.draw(event)

    def on_click(self, event):
        """Habndle mouse click events.

        Args:
            event (matplotlib event):
                The matplotlib event object

        Returns:
            None.

        Single left clicking adds new vertices to the list, double clicking adds vertices and finishes. Right
        clicking removes vertices.
        """
        if event.button == 1:  # Left mouse button
            if not event.dblclick:  # discard second click vertex addition.
                self.vertices.append((event.xdata, event.ydata))
            else:
                self.finished = True
        elif event.button == 3 and len(self.vertices) > 0:  # Right click toi remove last vertex
            del self.vertices[-1]
            if len(self.vertices) == 0:
                self.ov_layer.set_array(np.zeros(self.shape + (4,)))
            else:
                self.draw(event)

    def draw(self, event=None):
        """Handle the drawing of the selection shape.

        Keyword Args:
            event (matplotlib event):
                The matplotlib event object

        Returns:
            None.

        Delegates drawing to a draw_{shape} method.
        """
        if event is not None:
            vertices = np.array(self.vertices + [(event.xdata, event.ydata)])
        else:
            vertices = np.array(self.vertices)
        if vertices.size < 4:
            return
        vertices = np.round(vertices).astype(int)
        meth = getattr(self, f"draw_{self.obj}", lambda x: ([], []))
        rr, cc = meth(vertices)
        if self.invert:
            overlay = np.ones(self.shape + (4,))
            overlay[:, :] *= self.colour
            overlay[rr, cc] = [0, 0, 0, 0]
        else:
            overlay = np.zeros(self.shape + (4,))
            overlay[rr, cc] = self.colour

        # Add handles for the vertices.
        vertex_colour = np.copy(self.colour)
        vertex_colour[3] = 1.0  # Make vertices solid
        overlay[self.draw_vertices(vertices)] = vertex_colour

        self.ov_layer.set_array(overlay)

    def draw_vertices(self, vertices):
        """Return coordinates for small circules at each vertex."""
        radius = max(self.shape) / 50
        rr = np.array([], dtype=int)
        cc = np.array([], dtype=int)
        for vertex in vertices:
            vertex = (vertex[1], vertex[0])
            r, c = draw.disk(vertex, radius)
            rr = np.append(rr, r)
            cc = np.append(cc, c)
        return rr, cc

    def draw_poly(self, vertices):
        """Draw a polygon method using the specified vertices.

        Returns rr,cc coordinates.
        """
        if len(vertices) < 2:
            return ([], [])
        if len(vertices) == 2:
            rr, cc = draw.line(vertices[0, 1], vertices[0, 0], vertices[1, 1], vertices[1, 0])
        else:
            rr, cc = draw.polygon(vertices[:, 1], vertices[:, 0], self.shape)
        return rr, cc

    draw_poly.min_vertices = 3
    draw_poly.instructions = "Click to add vertices in order."

    def draw_circle(self, vertices):
        """Draw a circule or elipsoid."""
        if len(vertices) < 2:
            return ([], [])
        if len(vertices) == 2:  # Circle mode
            c0, r0 = vertices[0]
            c1, r1 = vertices[1]
            cc = (c0 + c1) // 2
            rc = (r0 + r1) // 2
            r = np.sqrt((rc - r1) ** 2 + (cc - c1) ** 2)
            return draw.disk((rc, cc), r, shape=self.shape)
        if len(vertices) == 3:
            p0 = vertices[0]
            p1 = vertices[1]
            p2 = vertices[2]
            bc = (np.dot(p0, p0) - np.dot(p1, p1)) / 2
            cd = (np.dot(p1, p1) - np.dot(p2, p2)) / 2
            det = (p0[0] - p1[0]) * (p1[1] - p2[1]) - (p1[0] - p2[0]) * (p0[1] - p1[1])
            if abs(det) < 1.0e-6:
                return ([], [])
            cx = (bc * (p1[1] - p2[1]) - cd * (p0[1] - p1[1])) / det
            cy = ((p0[0] - p1[0]) * cd - (p1[0] - p2[0]) * bc) / det
            r = np.sqrt((cx - p1[0]) ** 2 + (cy - p1[1]) ** 2)
            return draw.disk((cy, cx), r, shape=self.shape)
        if len(vertices) == 4:
            xc, yc = vertices.mean(axis=0)
            a, b = vertices.max(axis=0) - vertices.min(axis=0)
            x0 = [xc, yc, a, b]
            result = minimize(_straight_ellipse, x0=x0, args=(vertices,))
            xc, yc, a, b = result.x
            return draw.ellipse(yc, xc, b, a, shape=self.shape)
        if len(vertices) > 4:
            xc, yc = vertices.mean(axis=0)
            a, b = vertices.max(axis=0) - vertices.min(axis=0)
            phi = 0
            x0 = [xc, yc, a, b, phi]
            result = minimize(_rotated_ellipse, x0=x0, args=(vertices,))
            xc, yc, a, b, phi = result.x
            return draw.ellipse(yc, xc, b, a, shape=self.shape, rotation=phi)

    draw_circle.min_vertices = 2
    draw_circle.instructions = "2 or 3 perimeter vertices to define a circle\n4 or more to define an ellipse"

    def draw_rectangle(self, vertices):
        """Calculate the coordinates for a rectangle from the vertices."""
        if len(vertices) < 2:
            return ([], [])
        if len(vertices) == 2:
            c0, r0 = vertices[0]
            c1, r1 = vertices[1]
            rr, cc = draw.rectangle((r0, c0), (r1, c1), shape=self.shape)
        else:
            c0, r0 = vertices[-3]
            c1, r1 = vertices[-2]
            c2, r2 = vertices[-1]
            theta = np.arctan2(c1 - c0, r0 - r1)
            xp, yp = np.cos(theta), np.sin(theta)
            w = np.dot([c2 - c1, r2 - r1], [xp, yp])
            c2 = c1 + w * xp
            r2 = r1 + w * yp
            c3 = c0 + w * xp
            r3 = c0 + w * yp
            xvert = np.array([c0, c1, c2, c3], dtype=int)
            yvert = np.array([r0, r1, r2, r3], dtype=int)
            rr, cc = draw.polygon(yvert, xvert, shape=self.shape)
        return rr.astype(int), cc.astype(int)

    draw_rectangle.min_vertices = 2
    draw_rectangle.instructions = "Click to add corner vertices."

    def get_mask(self):
        """Convert a list of vertices to a mask array."""
        mask = np.ones(self.shape, dtype=bool) & self.invert
        vertices = np.array(self.vertices)
        meth = getattr(self, f"draw_{self.obj}", lambda x: ([], []))
        rr, cc = meth(vertices)
        mask[rr, cc] = ~self.invert
        return mask



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""Provides core functionality for the :py:class:`Stoner.Data` class."""

__all__ = [
    "metadataObject",
    "typeHintedDict",
    "regexpDict",
    "_setas",
    "DataArray",
    "array",
    "base",
    "interfaces",
    "methods",
    "operators",
    "property",
    "setas",
    "string_to_type",
    "exceptions",
    "utils",
    "Typing",
]

from .base import regexpDict, typeHintedDict, metadataObject, string_to_type
from .setas import setas as _setas
from .array import DataArray
from . import utils, exceptions, base, array
from . import Typing



"""Plot 3d fdata as a colourmap."""

# pylint: disable=invalid-name
import numpy as np


from Stoner import Data

x, y = np.meshgrid(np.linspace(-2, 2, 100), np.linspace(-2, 2, 100))
x = x.ravel()
y = y.ravel()
z = np.cos(4 * np.pi * np.sqrt(x**2 + y**2)) * np.exp(-np.sqrt(x**2 + y**2))

p = Data(np.column_stack((x, y, z)), column_headers=["X", "Y", "Z"])
p.setas = "xyz"

p.colormap_xyz()
p.title = "Colourmap plot"
